# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..memio import EUDByteReader

_global_parser = None


class _EUDParser(EUDByteReader):
    def __init__(self):
        super().__init__()
        self._number, self._digits = c.EUDCreateVariables(2)

    def _trim(self, b):
        lastblock = ut.EUDGetLastBlock()[1]
        cont, end = c.Forward(), c.Forward()

        trim = c.RawTrigger(actions=c.SetNextPtr(cont, end))
        for whitespace in " \t\n\f\r":
            c.RawTrigger(
                conditions=[b == ord(whitespace)],
                actions=self._digits.SetNumber(1 << 31),
            )
        cont << c.RawTrigger(
            conditions=[self._digits == 1 << 31],
            actions=[
                self._digits.SetNumber(0),
                c.SetNextPtr(cont, lastblock["loopstart"]),
            ],
        )
        end << c.NextTrigger()
        return trim

    def _sign(self, b, jumper):
        lastblock = ut.EUDGetLastBlock()[1]
        cont, end = c.Forward(), c.Forward()

        cs.DoActions(c.SetNextPtr(jumper, end), c.SetNextPtr(cont, end))
        c.RawTrigger(
            conditions=[b == ord("-")],
            actions=[
                self._digits.SetNumber(1 << 31),
                c.SetNextPtr(cont, lastblock["loopstart"]),
            ],
        )
        cont << c.RawTrigger(
            conditions=[b == ord("+")],
            actions=c.SetNextPtr(cont, lastblock["loopstart"]),
        )
        end << c.NextTrigger()

    @c.EUDMethod
    def _parse_dw(self):
        # \s*[+-]?\d+
        jumper, trim = c.Forward(), c.Forward()
        number, digits = self._number, self._digits
        cs.DoActions(
            number.SetNumber(0), digits.SetNumber(0), c.SetNextPtr(jumper, trim)
        )
        if cs.EUDInfLoop()():
            block = ut.EUDPeekBlock("infloopblock")[1]

            b = self.readbyte()
            jumper << c.RawTrigger()

            # Skip white space and pick up leading +/- sign if any.
            trim << self._trim(b)
            self._sign(b, jumper)

            # \d+
            is_digit = c.Forward()
            is_digit << c.RawTrigger(
                conditions=[b >= ord("0"), b <= ord("9")],
                actions=[
                    b.SubtractNumber(ord("0")),
                    c.SetNextPtr(is_digit, block["contpoint"]),
                ],
            )
            fin = c.NextTrigger()
            cs.EUDBreak()

            cs.EUDSetContinuePoint()
            # handle overflow
            cutoff, cutlim = divmod(0x7FFFFFFF, 10)
            c.RawTrigger(
                conditions=[
                    number == cutoff,
                    digits.AtLeastX(cutlim + 1, (1 << 31) - 1),
                ],
                actions=digits.SetNumberX(1 << 30, 1 << 30),
            )
            c.RawTrigger(
                conditions=[number >= cutoff + 1],
                actions=digits.SetNumberX(1 << 30, 1 << 30),
            )

            number *= 10
            # ignore leading zeros
            c.RawTrigger(
                conditions=[number == 0], actions=digits.SetNumberX(0, (1 << 30) - 1)
            )
            c.SeqCompute(
                [
                    (ut.EPD(is_digit) + 1, c.SetTo, fin),
                    (digits, c.Add, 1),
                    (number, c.Add, b),
                ]
            )
            c.RawTrigger(
                conditions=digits.AtLeastX(11, (1 << 30) - 1),
                actions=digits.SetNumberX(10, (1 << 30) - 1),
            )
        cs.EUDEndInfLoop()

        c.RawTrigger(
            conditions=digits.AtLeastX(1, 1 << 30),
            actions=number.SetNumber(0x7FFFFFFF),
        )
        if cs.EUDIf()(digits.AtLeastX(1, 1 << 31)):
            number << -number
        cs.EUDEndIf()
        cs.DoActions(digits.SetNumberX(0, 3 << 30))

        c.EUDReturn(number, digits)

    @c.EUDMethod
    def _parse_from_radix(self, radix):
        # \s*[+-]?(0[bB])?[01]+
        jumper, trim = c.Forward(), c.Forward()
        number, digits = self._number, self._digits
        cs.DoActions(
            number.SetNumber(0), digits.SetNumber(0), c.SetNextPtr(jumper, trim)
        )
        if cs.EUDInfLoop()():
            block = ut.EUDPeekBlock("infloopblock")[1]

            b = self.readbyte()
            jumper << c.RawTrigger()

            # Skip white space and pick up leading +/- sign if any.
            trim << self._trim(b)
            self._sign(b, jumper)

            # (0[bB])?
            is_zero, is_literal = c.Forward(), c.Forward()
            assume_decimal, cont, is_alphanumeric = (c.Forward() for _ in range(3))

            cs.DoActions(
                c.SetNextPtr(is_zero, assume_decimal),
                c.SetNextPtr(cont, is_alphanumeric),
            )
            is_zero << c.RawTrigger(
                nextptr=0,
                conditions=[b == ord("0")],
                actions=[
                    digits.AddNumber(1),
                    c.SetNextPtr(jumper, is_literal),
                    c.SetNextPtr(is_zero, block["loopstart"]),
                ],
            )
            # if there is a leading zero, check for binary literal
            is_literal << c.NextTrigger()
            for literal, base in zip("bBoOxX", (2, 2, 8, 8, 16, 16)):
                c.RawTrigger(
                    conditions=[b == ord(literal), radix == base],
                    actions=c.SetNextPtr(cont, block["loopstart"]),
                )
                c.RawTrigger(
                    conditions=[b == ord(literal), radix == 0],
                    actions=[
                        c.SetNextPtr(cont, block["loopstart"]),
                        radix.SetNumber(base),
                    ],
                )
            assume_decimal << c.RawTrigger(
                conditions=[radix == 0], actions=radix.SetNumber(10)
            )
            cutoff, cutlim = c.f_div(0xFFFFFFFF, radix)
            cont << c.RawTrigger(
                nextptr=0,
                actions=[
                    c.SetNextPtr(jumper, is_alphanumeric),
                    cutlim.AddNumber(1),
                ],
            )

            # [\da-zA-Z]+
            fin = c.Forward()
            is_alphanumeric << c.RawTrigger(
                actions=c.SetNextPtr(fin, block["loopend"])
            )
            c.RawTrigger(
                conditions=[b >= ord("0"), b <= ord("9")],
                actions=[
                    b.SubtractNumber(ord("0")),
                    c.SetNextPtr(fin, block["contpoint"]),
                ],
            )
            c.RawTrigger(
                conditions=[b >= ord("a"), b <= ord("z")],
                actions=[
                    b.SubtractNumber(ord("a") - 10),
                    c.SetNextPtr(fin, block["contpoint"]),
                ],
            )
            c.RawTrigger(
                conditions=[b >= ord("A"), b <= ord("Z")],
                actions=[
                    b.SubtractNumber(ord("A") - 10),
                    c.SetNextPtr(fin, block["contpoint"]),
                ],
            )
            fin << c.RawTrigger(nextptr=0)

            cs.EUDSetContinuePoint()
            cs.EUDBreakIf(b >= radix)

            # handle overflow
            c.RawTrigger(
                conditions=digits.AtLeastX(1, 1 << 30), actions=digits.AddNumber(-1)
            )
            if cs.EUDIf()(
                [number == cutoff, digits.AtLeastX(cutlim, (1 << 31) - 1)]
            ):
                cs.DoActions(digits.SetNumberX(1 << 30, 1 << 30))
            if cs.EUDElse()():
                cutoff += 1
                if cs.EUDIf()(number >= cutoff):
                    cs.DoActions(digits.SetNumberX(1 << 30, 1 << 30))
                cs.EUDEndIf()
                cutoff -= 1
            cs.EUDEndIf()

            number *= radix
            # ignore leading zeros
            c.RawTrigger(
                conditions=[number == 0], actions=digits.SetNumberX(0, (1 << 30) - 1)
            )
            c.SeqCompute([(digits, c.Add, 1), (number, c.Add, b)])
        cs.EUDEndInfLoop()

        c.RawTrigger(
            conditions=digits.AtLeastX(1, 1 << 30),
            actions=number.SetNumber(0x7FFFFFFF),
        )
        if cs.EUDIf()(digits.AtLeastX(1, 1 << 31)):
            number << -number
        cs.EUDEndIf()
        cs.DoActions(digits.SetNumberX(0, 3 << 30))

        c.EUDReturn(number, digits)


def _get_global_eudparser():
    global _global_parser
    if _global_parser is None:
        _global_parser = _EUDParser()
    return _global_parser


def f_parse(dst, radix=10):
    if radix in (int, 10):
        return _parse_dw(dst)
    else:
        return _parse_from_radix(dst, radix)


@c.EUDFunc
def _parse_dw(dst):
    global_parser = _get_global_eudparser()
    global_parser.seekoffset(dst)
    number, digits = global_parser._parse_dw()
    c.EUDReturn(number, digits)


@c.EUDFunc
def _parse_from_radix(dst, radix):
    global_parser = _get_global_eudparser()
    global_parser.seekoffset(dst)
    if cs.EUDIf()(radix == 10):
        number, digits = global_parser._parse_dw()
        c.EUDReturn(number, digits)
    cs.EUDEndIf()
    number, digits = global_parser._parse_from_radix(radix)
    c.EUDReturn(number, digits)



from helper import *
from eudplib.core.eudfunc import *


@EUDTracedFunc
def _t1(x):
    EUDTraceLog()
    if EUDIf()(x == 0):
        EUDTraceLog()
        EUDReturn(11)
    if EUDElseIf()(x == 1):
        EUDTraceLog()
        EUDReturn(12)
    EUDEndIf()
    EUDTraceLog()
    return 13


@EUDTracedFunc
def _t2():
    pass


@TestInstance
def test_trace():
    a = _t1(0)
    t1 = GetTraceStackDepth()
    b = _t1(1)
    t2 = GetTraceStackDepth()
    c = _t1(2)
    t3 = GetTraceStackDepth()
    test_equality(
        "EUDTracedFunc works well", [a, b, c, t1, t2, t3], [11, 12, 13, 0, 0, 0]
    )

    _t2()
    test_equality(
        "EUDTraceFunc for function with no returns", GetTraceStackDepth(), 0
    )


# phu54321/euddraft#3
EUDTracedTypedMethod



# Copyright 2024 by zzt (Defender).
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar

from .. import utils as ut
from ..core.rawtrigger.consttype import ConstType
from ..core.rawtrigger.strenc import EncodeUnit
from ..core.rawtrigger.typehint import Unit
from ..localize import _
from .offsetmap import (
    Bit1Member,
    BoolMember,
    ByteEnumMember,
    ByteMember,
    DwordEnumMember,
    DwordMember,
    EPDOffsetMap,
    Flag,
    FlingyMember,
    ImageMember,
    MapStringMember,
    MovementFlags,
    NotImplementedMember,
    PortraitMember,
    PositionMember,
    PositionXMember,
    PositionYMember,
    RankMember,
    RightClickActionMember,
    SfxDataMember,
    UnitMember,
    UnitOrderMember,
    UnitSizeMember,
    UpgradeMember,
    WeaponMember,
    WordEnumMember,
    WordMember,
)


class GroupFlags(ByteEnumMember):
    __slots__ = ()
    Zerg = Flag(0x01)
    """Uses underlings, can build on creep"""
    Terran = Flag(0x02)
    """Uses Supply, has sublabel, buildings will burn"""
    Protoss = Flag(0x04)
    """Uses Psi"""
    Men = Flag(0x08)
    Building = Flag(0x10)
    Factory = Flag(0x20)
    Independent = Flag(0x40)
    Neutral = Flag(0x80)


class BaseProperty(DwordEnumMember):
    """SpecialAbilityFlag in PyMS, UnitPrototypeFlags in bwapi, BaseProperty in GPTP"""  # noqa: E501

    __slots__ = ()
    Building = Flag(0x00000001)
    Addon = Flag(0x00000002)
    Flyer = Flag(0x00000004)
    Worker = Flag(0x00000008)
    "resource_miner in PyMS"
    Subunit = Flag(0x00000010)
    FlyingBuilding = Flag(0x00000020)
    Hero = Flag(0x00000040)
    RegeneratesHp = Flag(0x00000080)
    AnimatedIdle = Flag(0x00000100)
    Cloakable = Flag(0x00000200)
    TwoUnitsInOneEgg = Flag(0x00000400)
    SingleEntity = Flag(0x00000800)  # NeutralAccessories in GPTP
    "prevents multi-select, set on all pickup items"
    ResourceDepot = Flag(0x00001000)
    "Place where resources are brought back"
    ResourceContainer = Flag(0x00002000)
    "Resource Source. Can be selected by `ModifyUnitResourceAmount` action."
    Robotic = Flag(0x00004000)
    Detector = Flag(0x00008000)
    Organic = Flag(0x00010000)
    RequiresCreep = Flag(0x00020000)
    Unused = Flag(0x00040000)
    RequiresPsi = Flag(0x00080000)
    Burrowable = Flag(0x00100000)
    Spellcaster = Flag(0x00200000)
    "Can be selected by `ModifyUnitEnergy` action"
    PermanentCloak = Flag(0x00400000)
    PickupItem = Flag(0x00800000)
    "data disc, crystals, mineral chunks, gas tanks, etc."
    IgnoresSupplyCheck = Flag(0x01000000)
    "MorphFromOtherUnit in GPTP"
    MediumOverlay = Flag(0x02000000)
    "Used to determine overlay for various spells and effects"
    LargeOverlay = Flag(0x04000000)
    AutoAttackAndMove = Flag(0x08000000)
    "battle_reactions in PyMS"
    CanAttack = Flag(0x10000000)
    "full_auto_attack in PyMS"
    Invincible = Flag(0x20000000)
    Mechanical = Flag(0x40000000)
    ProducesUnits = Flag(0x80000000)
    "It can produce units directly (making buildings doesn't count)"


class AvailabilityFlags(WordEnumMember):
    __slots__ = ()
    NonNeutral = Flag(0x001)
    UnitListing = Flag(0x002)
    "set availability to be created by CreateUnit action"
    MissionBriefing = Flag(0x004)
    PlayerSettings = Flag(0x008)
    AllRaces = Flag(0x010)
    SetDoodadState = Flag(0x020)
    NonLocationTriggers = Flag(0x040)
    UnitHeroSettings = Flag(0x080)
    LocationTriggers = Flag(0x100)
    BroodWarOnly = Flag(0x200)


class TrgUnit(EPDOffsetMap, ConstType):
    __slots__ = ()
    flingy: ClassVar[FlingyMember] = FlingyMember("array", 0x6644F8)
    "Flingy: Flingy of unit. 유닛의 비행정보."
    subUnit: ClassVar[UnitMember] = UnitMember("array", 0x6607C0)
    "TrgUnit: Subunit of unit. 유닛의 부가유닛"
    # subunit2: ClassVar[WordMember] = WordMember("array", 0x660C38)
    # subunit2 is unused. 유닛의 두번째 부가유닛. 사용되지 않습니다.
    # infestationUnit is not implemented yet (different beginning index)
    # SCBW_DATA(u16*, InfestedUnitPartial, unitsDat[3].address);
    # 0x664980, (Id - UnitId::TerranCommandCenter) for it to work,
    # last valid id is UnitId::Special_OvermindCocoon
    # 퀸이 감염 시켰을 경우 바뀌는 대상 유닛. 106-201 (건물)에만 존재합니다.
    constructionGraphic: ClassVar[ImageMember] = ImageMember(
        "array", 0x6610B0, stride=4
    )
    """Image: Sets the graphic when the unit is built.

    유닛이 건설될 때의 그래픽을 설정합니다.
    """
    startDirection: ClassVar[ByteMember] = ByteMember("array", 0x6605F0)
    """u8: Direction unit will face after it is created.

    Values start at 0 (the unit will face the top of the screen) and go on clockwise
    through subsequent turning stages until 31 (unit will face a little left from the
    complete turn). Value of 32 means unit will face a random direction.

    유닛이 생성된 후 향하게 될 방향입니다. 값은 0(위쪽)에서 시작하여 이후 시계
    방향으로 31(위쪽에서 약간 왼쪽)까지 증가합니다. 값이 32이면 랜덤 방향을 향하게
    됩니다.
    """
    hasShield: ClassVar[BoolMember] = BoolMember("array", 0x6647B0)
    """bool: Flag indicating if the unit's shields are enabled.

    When enabled, the unit's shields can be modified with `ModifyUnitShields` action.
    Note that Terran and Zerg buildings under construction do not regenerate shields.

    활성화하면 `ModifyUnitShields` 액션으로 유닛의 보호막을 수정할 수 있습니다. 건설
    중인 테란 및 저그 건물은 보호막을 재생하지 않습니다.
    """
    maxShield: ClassVar[WordMember] = WordMember("array", 0x660E00)
    "u16: Shield amount of unit. 유닛의 최대 실드량."
    maxHp: ClassVar[DwordMember] = DwordMember("array", 0x662350)
    """i32: Max hit point of unit.

    Values above '10000' are allowed in the game, but no numbers are displayed.

    유닛의 체력. '10000' 이상의 값은 게임에서 허용되지만 수치가 표시되지 않습니다.
    """
    elevation: ClassVar[ByteMember] = ByteMember("array", 0x663150)
    """u8: The unit's visible height.

    When set to Air (12 <= elevation <= 18), allows the unit to pass over impassable
    terrain or other ground units. The higher the value, sorts sprite elevation
    higher and appear above other units.

    유닛의 표시 높이. 공중으로 설정할 경우 (12 <= 높이 <= 18) 통과 불가 지형이나 다른
    지상 유닛 위로 지나다닐 수 있습니다. 값이 높을수록 지형이나 다른 유닛보다 위에
    표시됩니다.
    """
    movementFlags = MovementFlags("array", 0x660FC8)
    rank: ClassVar[RankMember] = RankMember("array", 0x663DD0)
    """u8: Stat text id offset for unit rank, starting from [1302] Recruit.

    Sets the rank of the unit. Unit status text is prioritized in the following order
    (1) Hallucination (2) Buildings that are Disabled, Lockdown, Stasis, or Maelstrom
    (3) Acid Spores (4) Blind (5) Parasite (6) Detector (7) Rank. For a rank to be
    visible, it must have no `Building` and `SingleEntity` flags in `baseProperty`,
    and not be a Civilian or Spider Mine. Gantrithor, Battlecruiser, Norad II, and
    Hyperion must use their default names (`nameString` must be 0) to use the
    Tassadar/Admiral rank as fixed. Other units must only have `Terran` enabled in
    their race flags in `groupFlags` to see their ranks.

    유닛의 계급을 설정합니다. 유닛 상태 텍스트의 우선순위는 다음과 같습니다:
    (1) 할루시네이션 (2) 디시블, 락다운, 스테이시스, 마엘스트롬에 걸린 건물
    (3) 애시드 스포어 (4) 블라인드 (5) 패러사이트 (6) 디텍터 (7) 계급. 계급이
    보이려면 `baseProperty`에 `Building`과 `SingleEntity` 플래그가 없어야하고,
    시민이나 스파이더 마인이 아니어야 합니다. 간트리서, 배틀크루저, 노라드 II,
    히페리온은 기본 이름을 사용해야 (`nameString`이 0이어야) 태사다르/제독 계급을
    고정으로 사용합니다. 이 외의 유닛은 `groupFlags`의 종족 플래그에서 `Terran`만
    활성화해야 계급이 보입니다.
    """
    computerIdleOrder: ClassVar[UnitOrderMember] = UnitOrderMember("array", 0x662EA0)
    """UnitOrder: An order issued when unit owned by computer is doing nothing.

    컴퓨터의 유닛이 아무것도 하고있지 않을 때 내려지는 명령.
    """
    humanIdleOrder: ClassVar[UnitOrderMember] = UnitOrderMember("array", 0x662268)
    """UnitOrder: An order issued when unit owned by huamn is doing nothing

    사람의 유닛이 아무것도 하고있지 않을 때 내려지는 명령.
    """
    returnToIdleOrder: ClassVar[UnitOrderMember] = UnitOrderMember("array", 0x664898)
    """UnitOrder: An order issued when unit return to idle state.

    유닛이 이전 명령을 수행한 후 아무것도 하지 않는 상태로 돌아가는 명령.
    """
    attackUnitOrder: ClassVar[UnitOrderMember] = UnitOrderMember("array", 0x663320)
    """UnitOrder: An order performed when a unit is ordered to attack.

    Also performed by right-clicking if the target is an enemy.

    유닛이 대상을 공격하라고 명령 받았을 때 수행하는 명령. 대상이 적이면 우클릭으로도
    수행합니다.
    """
    attackMoveOrder: ClassVar[UnitOrderMember] = UnitOrderMember("array", 0x663A50)
    groundWeapon: ClassVar[WeaponMember] = WeaponMember("array", 0x6636B8)
    airWeapon: ClassVar[WeaponMember] = WeaponMember("array", 0x6616E0)
    maxGroundHits: ClassVar[ByteMember] = ByteMember("array", 0x6645E0)
    """u8: (UNUSED) Maximum number of hits this unit can deal to ground targets.

    This attribute represents the maximum number of hits this unit can inflict on a
    target with its ground weapon. For example, the Psi Blades have a `damageFactor`
    of 1, while the Zealot has a `maxGroundHits` of 2. The actual number of attacks
    is determined by the iscript.

    유닛이 지상 무기로 대상을 타격할 수 있는 최대 명중 횟수입니다. 예를 들어,
    사이오닉 검의 `damageFactor`는 1이고 질럿의 `maxGroundHits`은 2입니다.
    실제 공격 횟수는 iscript에서 결정됩니다.
    """
    maxAirHits: ClassVar[ByteMember] = ByteMember("array", 0x65FC18)
    """u8: (UNUSED) Maximum number of hits this unit can deal to air targets.

    This attribute represents the maximum number of hits this unit can inflict on a
    target with its air weapon. For example, Halo Rockets have a `damageFactor` of 2,
    while the Valkyrie has a `maxAirHits` of 4. The actual number of attacks is
    determined by the iscript.

    유닛이 공중 무기로 대상을 타격할 수 있는 최대 명중 횟수입니다. 예를 들어, 광륜
    로켓의 `damageFactor`는 최대치인 2이고 발키리의 `maxAirHits`은 4입니다. 실제 공격
    횟수는 iscript에서 결정됩니다.
    """
    ignoreStrategicSuicideMissions: ClassVar[BoolMember] = BoolMember(
        "array", 0x660178
    )
    """bool: Flag indicating if the unit is excluded from Strategic Suicide Missions.

    When `ignoreStrategicSuicideMissions` is enabled, the unit will be excluded from
    Strategic Suicide Missions. Disabling this flag does not have a noticeable effect
    . By default, `ignoreStrategicSuicideMissions` and `dontBecomeGuard` are either
    both enabled or both disabled for every unit.

    `ignoreStrategicSuicideMissions`가 활성화되면 해당 유닛이 Strategic Suicide
    Missions에서 제외됩니다. 이 플래그를 비활성화해도 눈에 띄는 효과는 없습니다.
    기본적으로 모든 유닛에서 `ignoreStrategicSuicideMissions`와 `dontBecomeGuard`는
    둘 다 켜져있거나 둘 다 꺼져있습니다.
    """
    dontBecomeGuard: ClassVar[Bit1Member] = Bit1Member("array", 0x660178)
    """bool: Flag to prevent unit from returning to original position.

    Enabling `dontBecomeGuard` can cause game crashes, especially if the CPU controls
    the unit. This flag prevents the unit from returning to its original position if
    the target disappears, but does not affect units that already exist. It primarily
    impacts units with AI ptr type 1 or 4, which are subject to Strategic Suicide
    Missions. Units with this flag set will only be assigned AI if they have
    `baseProperty.Worker` (becoming unitAI type 2), or if they have
    `baseProperty.Building` and are not geysers, or are larva/egg/overlord (becoming
    unitAI type 3).

    `dontBecomeGuard`를 활성화하면 특히 CPU가 유닛을 제어할 때 게임이 크래시할 수
    있습니다. 이 플래그는 명령 대상이 사라졌을 때 유닛이 원래 위치로 돌아가는 것을
    방지합니다. 이미 존재하는 유닛에는 영향을 주지 않습니다. 이 플래그는 주로 AI ptr
    타입 1 또는 4가 있는 유닛에 영향을 미치며, 이러한 유닛은 Strategic Suicide
    Missions에 대상이 됩니다. 이 플래그가 설정된 유닛은 `baseProperty.Worker`가
    있거나 (유닛 AI 타입 2가 됨), `baseProperty.Building`이 있고 베스핀 간헐천이
    아니거나, 라바/에그/오버로드인 경우 (유닛AI 타입 3이 됨) AI가 할당됩니다.
    """
    baseProperty = BaseProperty("array", 0x664080)
    seekRange: ClassVar[ByteMember] = ByteMember("array", 0x662DB8)
    sightRange: ClassVar[ByteMember] = ByteMember("array", 0x663238)
    armorUpgrade: ClassVar[UpgradeMember] = UpgradeMember("array", 0x6635D0)
    sizeType: ClassVar[UnitSizeMember] = UnitSizeMember("array", 0x662180)
    """u8: Size classification of the unit.

    Defines the size of unit as one of four types: `"Small"`, `"Medium"`, `"Large"`,
    or `"Independent"`. This size affects the damage calculation for various damage
    types. Damage multipliers for weapon damage types against each unit size are
    located at memory address `0x515B84`.

    유닛의 크기를 네 가지 타입 중 하나로 정의합니다: `"Small"`, `"Medium"`,
    `"Large"`, `"Independent"`. 방어 타입은 무기 타입에 따른 대미지 계산에 영향을
    줍니다. 각 무기 타입마다 방어 타입에 대한 대미지 비율은 메모리 주소 `0x515B84`에
    있습니다.
    """
    armor: ClassVar[ByteMember] = ByteMember("array", 0x65FEC8)
    rightClickAction: ClassVar[RightClickActionMember] = RightClickActionMember(
        "array", 0x662098
    )
    readySound: ClassVar[SfxDataMember] = SfxDataMember("array", 0x661FC0)
    whatSoundStart: ClassVar[SfxDataMember] = SfxDataMember("array", 0x65FFB0)
    whatSoundEnd: ClassVar[SfxDataMember] = SfxDataMember("array", 0x662BF0)
    pissedSoundStart: ClassVar[SfxDataMember] = SfxDataMember("array", 0x663B38)
    pissedSoundEnd: ClassVar[SfxDataMember] = SfxDataMember("array", 0x661EE8)
    yesSoundStart: ClassVar[SfxDataMember] = SfxDataMember("array", 0x663C10)
    yesSoundEnd: ClassVar[SfxDataMember] = SfxDataMember("array", 0x661440)
    buildingDimensions: ClassVar[PositionMember] = PositionMember("array", 0x662860)
    """Position: Dimensions for building placement and visibility.

    This dimension is used when determining if units with the Building flag can fit
    in the available spaace. Units without the Building flag will rely on their
    collision dimensions instead. Setting to 0x0 will make the unit invisible: It
    won't appear on screen and minimap, can't be selected by the mouse, and can't be
    targetted by other units. `Bring` condition will not locate it. Setting this to
    a size of 31x31 or smaller will allow buildings to be built on any terrain,
    including water and cliffs, although the placement mechanics are a little wonky.

    건물 플래그가 있는 유닛이 공간에 들어갈 수 있는지 여부를 결정할 때 사용됩니다.
    건물 플래그가 없는 유닛은 대신 유닛 크기에 의존합니다. 0x0으로 설정하면 유닛이
    보이지 않게 됩니다: 화면과 미니맵에 나타나지 않고 마우스로 선택할 수 없으며 다른
    유닛이 타겟팅할 수 없습니다. `Bring` 조건은 유닛을 찾지 못합니다. 이 크기를 31x31
    이하로 설정하면 배치 메커니즘이 약간 불안정하지만 물과 절벽을 포함한 모든 지형에
    건물을 지을 수 있습니다.
    """
    addonPlacement: ClassVar[NotImplementedMember] = NotImplementedMember(
        "array", 0x6626E0
    )  # Position
    """AddonPlacement is not implemented yet because its beginning index isn't 0."""
    unitBoundsLT: ClassVar[PositionMember] = PositionMember(
        "array", 0x6617C8, stride=8
    )
    unitBoundsRB: ClassVar[PositionMember] = PositionMember(
        "array", 0x6617CC, stride=8
    )
    unitBoundsL: ClassVar[PositionXMember] = PositionXMember(
        "array", 0x6617C8, stride=8
    )
    unitBoundsT: ClassVar[PositionYMember] = PositionYMember(
        "array", 0x6617CA, stride=8
    )
    unitBoundsR: ClassVar[PositionXMember] = PositionXMember(
        "array", 0x6617CC, stride=8
    )
    unitBoundsB: ClassVar[PositionYMember] = PositionYMember(
        "array", 0x6617CE, stride=8
    )
    portrait: ClassVar[PortraitMember] = PortraitMember("array", 0x662F88)
    mineralCost: ClassVar[WordMember] = WordMember("array", 0x663888)
    gasCost: ClassVar[WordMember] = WordMember("array", 0x65FD00)
    timeCost: ClassVar[WordMember] = WordMember("array", 0x660428)
    requirementOffset: ClassVar[WordMember] = WordMember("array", 0x660A70)
    groupFlags = GroupFlags("array", 0x6637A0)
    supplyProvided: ClassVar[ByteMember] = ByteMember("array", 0x6646C8)
    supplyUsed: ClassVar[ByteMember] = ByteMember("array", 0x663CE8)
    transportSpaceProvided: ClassVar[ByteMember] = ByteMember("array", 0x660988)
    transportSpaceRequired: ClassVar[ByteMember] = ByteMember("array", 0x664410)
    buildScore: ClassVar[WordMember] = WordMember("array", 0x663408)
    killScore: ClassVar[WordMember] = WordMember("array", 0x663EB8)
    nameString: ClassVar[MapStringMember] = MapStringMember("array", 0x660260)
    """TrgString: Map string id for the unit's name.

    When this property is non-zero, the unit's name is read from the strings in
    the currently loaded map (CHK) rather than from the `stat_txt.tbl` file.

    이 속성이 0이 아닌 경우, 유닛의 이름은 `stat_txt.tbl` 파일 대신 현재 로드된
    맵(CHK) 내의 문자열에서 읽어옵니다.
    """
    broodWarFlag: ClassVar[ByteMember] = ByteMember("array", 0x6606D8)  # bool?
    availabilityFlags = AvailabilityFlags("array", 0x661518)

    @ut.classproperty
    def range(self):
        return (0, 227, 1)

    @classmethod
    def cast(cls, _from: Unit):
        if isinstance(_from, ConstType) and not isinstance(_from, cls):
            raise ut.EPError(_('"{}" is not a {}').format(_from, cls.__name__))
        return super().cast(_from)

    def __init__(self, initval: Unit) -> None:
        super().__init__(EncodeUnit(initval))



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .cpprint import (
    FixedText,
    PColor,
    PName,
    f_cpstr_print,
    f_eprintln,
    f_gettextptr,
    f_raise_CCMU,
)
from .cpstr import GetMapStringAddr
from .cputf8 import f_cp949_to_utf8_cpy
from .eudprint import f_dbstr_print
from .fmtprint import f_eprintAll, f_eprintf, f_sprintf, f_sprintf_cp
from .locale import LocalLocale
from .parse import f_parse
from .pname import IsPName, SetPName, SetPNamef
from .strall import DisplayTextAll, DisplayTextAllAt, f_printAll, f_printAllAt
from .strbuffer import (
    DisplayTextAt,
    GetGlobalStringBuffer,
    StringBuffer,
    f_printAt,
    f_println,
    f_simpleprint,
)
from .strcommon import epd2s, hptr, ptr2s
from .strfunc import f_strcmp, f_strcpy, f_strlen, f_strlen_epd, f_strnstr
from .tblprint import (
    GetTBLAddr,
    f_eprintln2,
    f_settbl,
    f_settbl2,
    f_settblf,
    f_settblf2,
)
from .texteffect import TextFX_FadeIn, TextFX_FadeOut, TextFX_Remove, TextFX_SetTimer

__all__ = [
    "FixedText",
    "PColor",
    "PName",
    "f_cpstr_print",
    "f_eprintln",
    "f_gettextptr",
    "f_raise_CCMU",
    "GetMapStringAddr",
    "f_cp949_to_utf8_cpy",
    "f_dbstr_print",
    "f_eprintAll",
    "f_eprintf",
    "f_sprintf",
    "f_sprintf_cp",
    "LocalLocale",
    "f_parse",
    "IsPName",
    "SetPName",
    "SetPNamef",
    "DisplayTextAll",
    "DisplayTextAllAt",
    "f_printAll",
    "f_printAllAt",
    "DisplayTextAt",
    "GetGlobalStringBuffer",
    "StringBuffer",
    "f_printAt",
    "f_println",
    "f_simpleprint",
    "epd2s",
    "hptr",
    "ptr2s",
    "f_strcmp",
    "f_strcpy",
    "f_strlen",
    "f_strlen_epd",
    "f_strnstr",
    "GetTBLAddr",
    "f_eprintln2",
    "f_settbl",
    "f_settbl2",
    "f_settblf",
    "f_settblf2",
    "TextFX_FadeIn",
    "TextFX_FadeOut",
    "TextFX_Remove",
    "TextFX_SetTimer",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
import os
from tempfile import NamedTemporaryFile

from .. import utils as ut
from ..bindings._rust import mpqapi
from ..localize import _

_addedFiles: dict[bytes, tuple[str, str | bytes | bytearray | None, bool]] = {}  # noqa: N816


def update_filelist_by_listfile(mpqr: mpqapi.MPQ) -> None:
    """Use listfile to get list of already existing files."""

    _addedFiles.clear()
    try:
        flist = mpqr.get_file_names_from_listfile()
    # no listfile -> add default listfile
    except Exception:
        flist = ["staredit\\scenario.chk", "(listfile)"]

    if not flist:
        flist = ["staredit\\scenario.chk", "(listfile)"]

    for fname in flist:
        MPQAddFile(fname, None)


def MPQCheckFile(fname: str) -> bytes:  # noqa: N802
    """Check if filename is already exist.

    :param fname: Desired filename in mpq
    """

    # make fname case_insensitive
    from ntpath import normcase, normpath

    fname_key = ut.u2b(normpath(normcase(fname)))

    ut.ep_assert(
        fname_key not in _addedFiles,
        _('MPQ filename duplicate : "{}"').format(fname),
    )

    return fname_key


def MPQAddFile(  # noqa: N802
    fname: str, path_or_content: str | bytes | bytearray | None, is_wav: bool = False
) -> None:
    """Add file/wave to output map.

    :param fname: Desired filename in mpq
    :param path_or_content: File path or content to put inside.
    :param is_wav: Is file wave file? Wave file can be lossy compressed if this
        flag is set to True.

    .. note::
        This function may error if duplicate filenames is found. However, not
        all duplicated filenames are guaranteed to be catched here. Some of
        them may be catched at _update_mpq(internal) function.
    """

    if not (
        isinstance(path_or_content, str | bytes | bytearray)
        or path_or_content is None
    ):
        raise ut.EPError(
            _("Invalid file path or content: {}").format(path_or_content)
        )

    # make fname case_insensitive
    fname_key = MPQCheckFile(fname)

    _addedFiles[fname_key] = (fname, path_or_content, is_wav)


def MPQAddWave(fname: str, path_or_content: str | bytes | bytearray | None) -> None:  # noqa: N802
    """Add wave to output map.

    :param fname: Desired filename in mpq
    :param path_or_content: File path or content to put inside.

    .. note:: See `MPQAddFile` for more info
    """
    MPQAddFile(fname, path_or_content, True)


def _update_mpq(mpqw: mpqapi.MPQ) -> None:
    """Really append additional mpq file to mpq file.

    `MPQAddFile` queues addition, and _update_mpq really adds them.
    """

    count = mpqw.get_max_file_count()
    if count < len(_addedFiles):  # need to increase max file count
        max_count = max(1024, 1 << (len(_addedFiles)).bit_length())
        try:
            mpqw.set_max_file_count(max_count)
        except Exception as e:
            raise ut.EPError(_("Failed to increase max file limit")) from e

    for fname, path_or_content, is_wav in _addedFiles.values():
        if path_or_content is None:
            continue

        if isinstance(path_or_content, str):
            file_path = path_or_content
        else:
            # Create temporary file
            f = NamedTemporaryFile(delete=False)
            f.write(bytes(path_or_content))
            file_path = f.name
            f.close()

        try:
            mpqw.add_file(fname, file_path)
        except Exception as e:
            raise ut.EPError(
                _("Failed adding file {} to mpq: May be duplicate").format(fname)
            ) from e

        if not isinstance(path_or_content, str):
            os.unlink(file_path)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import Self

from ... import utils as ut
from ...localize import _
from ..allocator import ConstExpr, IsConstExpr
from .consttype import Byte, Dword, Word

if TYPE_CHECKING:
    from ..allocator.payload import ObjCollector, RlocInt_C, _PayloadBuffer
    from .rawtriggerdef import RawTrigger

_acttypes: dict[int, str] = {
    0: "(no action)",
    1: "Victory",
    2: "Defeat",
    3: "PreserveTrigger",
    4: "Wait",
    5: "PauseGame",
    6: "UnpauseGame",
    7: "Transmission",
    8: "PlayWAV",
    9: "DisplayText",
    10: "CenterView",
    11: "CreateUnitWithProperties",
    12: "SetMissionObjectives",
    13: "SetSwitch",
    14: "SetCountdownTimer",
    15: "RunAIScript",
    16: "RunAIScriptAt",
    17: "LeaderBoardControl",
    18: "LeaderBoardControlAt",
    19: "LeaderBoardResources",
    20: "LeaderBoardKills",
    21: "LeaderBoardScore",
    22: "KillUnit",
    23: "KillUnitAt",
    24: "RemoveUnit",
    25: "RemoveUnitAt",
    26: "SetResources",
    27: "SetScore",
    28: "MinimapPing",
    29: "TalkingPortrait",
    30: "MuteUnitSpeech",
    31: "UnMuteUnitSpeech",
    32: "LeaderboardComputerPlayers",
    33: "LeaderboardGoalControl",
    34: "LeaderboardGoalControlAt",
    35: "LeaderboardGoalResources",
    36: "LeaderboardGoalKills",
    37: "LeaderboardGoalScore",
    38: "MoveLocation",
    39: "MoveUnit",
    40: "LeaderboardGreed",
    41: "SetNextScenario",
    42: "SetDoodadState",
    43: "SetInvincibility",
    44: "CreateUnit",
    45: "SetDeaths",
    46: "Order",
    47: "Comment",
    48: "GiveUnits",
    49: "ModifyUnitHitPoints",
    50: "ModifyUnitEnergy",
    51: "ModifyUnitShields",
    52: "ModifyUnitResourceAmount",
    53: "ModifyUnitHangarCount",
    54: "PauseTimer",
    55: "UnpauseTimer",
    56: "Draw",
    57: "SetAllianceStatus",
}


class Action(ConstExpr):

    """Action class.

    Memory layout.

     ======  ============= ========  ==========
     Offset  Field Name    Position  EPD Player
     ======  ============= ========  ==========
       +00   locid1         dword0   EPD(act)+0
       +04   strid          dword1   EPD(act)+1
       +08   wavid          dword2   EPD(act)+2
       +0C   time           dword3   EPD(act)+3
       +10   player1        dword4   EPD(act)+4
       +14   player2        dword5   EPD(act)+5
       +18   unitid         dword6   EPD(act)+6
       +1A   acttype
       +1B   amount
       +1C   flags          dword7   EPD(act)+7
       +1D   internal[3
     ======  ============= ========  ==========
    """

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls, None)

    def __init__(
        self,
        locid1: Dword,
        strid: Dword,
        wavid: Dword,
        time: Dword,
        player1: Dword,
        player2: Dword,
        unitid: Word,
        acttype: Byte,
        amount: Byte,
        flags: Byte,
        *,
        eudx: Word = 0,
    ) -> None:
        """See :mod:`eudplib.base.stocktrg` for stock actions list."""
        super().__init__()
        self.fields: list[Dword] = [
            locid1,
            strid,
            wavid,
            time,
            player1,
            player2,
            unitid,
            acttype,
            amount,
            flags,
            0,
            eudx,
        ]
        self.parenttrg: RawTrigger | None = None
        self.actindex: int | None = None

    def __copy__(self) -> Action:
        return self.__class__(*self.fields[:10], eudx=self.fields[11])  # type: ignore[arg-type]

    def disable(self) -> None:
        if isinstance(self.fields[9], ConstExpr):
            raise ut.EPError(_("Can't disable non-const Action flags"))
        self.fields[9] |= 2

    # -------

    def _invalid_action(self, i: int) -> str:
        acttype = self.fields[7]
        actname = _acttypes[acttype] if isinstance(acttype, int) else acttype
        return _("Invalid fields for action{} {}:").format(i, actname)

    def CheckArgs(self, i: int) -> None:  # noqa: N802
        fields = self.fields
        if (
            IsConstExpr(fields[0])
            and IsConstExpr(fields[1])
            and IsConstExpr(fields[2])
            and IsConstExpr(fields[3])
            and IsConstExpr(fields[4])
            and IsConstExpr(fields[5])
            and isinstance(fields[6], int)
            and isinstance(fields[7], int)
            and isinstance(fields[8], int)
            and isinstance(fields[9], int)
        ):
            return

        error = [self._invalid_action(i)]
        fieldname = [
            "location",
            "string",
            "wav_string",
            "time",
            "player",
            "amount",
            "unit_type",
            "action_type",
            "modifier",
            "flags",
            "padding",
            "maskflag",
        ]
        acttype = fields[7]
        if isinstance(acttype, int):
            if acttype == 45:  # SetDeaths
                fieldname[0] = "bitmask"
            elif acttype == 48:  # GiveUnits
                fieldname[4] = "giver"
                fieldname[5] = "taker"
            elif acttype == 38:  # MoveLocation
                fieldname[0] = "move_location"
                fieldname[5] = "dest_location"
            elif acttype in (39, 46):  # MoveUnit, Order
                fieldname[0] = "start_location"
                fieldname[5] = "dest_location"
            elif acttype == 11:  # CreateUnitwithProperties
                fieldname[5] = "unit_properties"
            elif acttype in (15, 16):  # RunAIScript(At)
                fieldname[5] = "aiscript"
            elif acttype == 13:  #  SetSwitch
                fieldname[5] = "switch_id"

            if acttype in (21, 27, 37):  # LeaderBoard (Score),  SetScore
                fieldname[6] = "score_type"
            elif acttype in (19, 26, 35):
                fieldname[6] = "resource_type"
            elif acttype == 57:
                fieldname[6] = "ally_status"

            if acttype in (23, 25, 39, 48, 49, 50, 51, 52, 53):
                fieldname[8] = "count"
            elif acttype in (32, 42, 43):
                fieldname[8] = "propstate"
            elif acttype == 13:
                fieldname[8] = "switch_action"
            elif acttype == 46:
                fieldname[8] = "unit_order"

        for i, field in enumerate(fields):
            if (i < 6 and not IsConstExpr(field)) or (
                i >= 6 and not isinstance(field, int)
            ):
                error.append(
                    "\t" + _("invalid {}: {}").format(fieldname[i], repr(field))
                )

        raise ut.EPError("\n".join(error))

    def SetParentTrigger(self, trg: RawTrigger, index: int) -> None:  # noqa: N802
        if self.parenttrg is not None:
            raise ut.EPError(_("Actions cannot be shared by two triggers."))
        if trg is None:
            raise ut.EPError(_("Trigger should not be null."))
        if not (0 <= index < 64):
            raise ut.EPError(_("Triggers out of range"))

        self.parenttrg = trg
        self.actindex = index

    def Evaluate(self) -> RlocInt_C:  # noqa: N802
        if self.parenttrg is None or self.actindex is None:
            # fmt: off
            err = _("Orphan action. This often happens when you try to do arithmetics with actions.")  # noqa: E501
            # fmt: on
            raise ut.EPError(err)
        return self.parenttrg.Evaluate() + 8 + 320 + 32 * self.actindex

    def CollectDependency(self, pbuffer: ObjCollector) -> None:  # noqa: N802
        for field in self.fields[:6]:
            if not isinstance(field, int):
                pbuffer.WriteDword(field)  # type: ignore[arg-type]

    def WritePayload(self, pbuffer: _PayloadBuffer) -> None:  # noqa: N802
        pbuffer.WritePack("IIIIIIHBBBBH", self.fields)  # type: ignore[arg-type]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...utils import EPD
from .condition import Condition
from .constenc import (
    EncodeComparison,
    EncodePlayer,
    EncodeResource,
    EncodeScore,
    EncodeSwitchState,
    Kills,  # for __calls__ binding
)
from .consttype import Dword
from .strenc import EncodeLocation, EncodeSwitch, EncodeUnit
from .typehint import (
    Comparison,
    Location,
    Player,
    Resource,
    SwitchState,
    Unit,
    _Score,
    _Switch,
)


def CountdownTimer(comparison: Comparison, time: Dword) -> Condition:  # noqa: N802
    """Checks countdown timer.

    Example::

        CountdownTimer(AtLeast, 10)

    Memory Layout::

        0000 0000 0000 0000 TTTT TTTT 0000 CP01 0000

        T : time, CP : Comparison.
    """
    comparison = EncodeComparison(comparison)
    return Condition(0, 0, time, 0, comparison, 1, 0, 0)


def Command(  # noqa: N802
    player: Player, comparison: Comparison, number: Dword, unit: Unit
) -> Condition:
    """[Player] commands [Comparison] [Number] [Unit].

    Example::
        Command(Player1, AtLeast, 30, "Terran Marine")


    """
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(0, player, number, unit, comparison, 2, 0, 0)


def Bring(  # noqa: N802
    player: Player,
    comparison: Comparison,
    number: Dword,
    unit: Unit,
    location: Location,
) -> Condition:
    """Player brings quantity units to location.

    This states that a player is required to bring 'X' number of units to a
    specific location. The units can be any player-controlled unit available
    in the game.
    """
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    return Condition(location, player, number, unit, comparison, 3, 0, 0)


def Accumulate(  # noqa: N802
    player: Player,
    comparison: Comparison,
    number: Dword,
    resource_type: Resource,
) -> Condition:
    """Player accumulates quantity resources.

    Accumulate requires that the player gather enough of a specific resource.
    """
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    resource_type = EncodeResource(resource_type)
    return Condition(0, player, number, 0, comparison, 4, resource_type, 0)


# 'Kills' is already defined inside constenc, so we just add __call__ method
# to there instead of creating new function
def __kills__internal(
    player: Player, comparison: Comparison, number: Dword, unit: Unit
) -> Condition:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(0, player, number, unit, comparison, 5, 0, 0)


Kills._internalf = __kills__internal


def CommandMost(unit: Unit) -> Condition:  # noqa: N802
    """Current player commands the most units.

    Command the Most requires that you command the most of the defined units.
    These units can be any player-controlled unit available in the game.
    This condition compares all players in the game, including neutral and
    rescuable units.
    """
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 6, 0, 0)


def CommandMostAt(unit: Unit, location: Location) -> Condition:  # noqa: N802
    """Current player commands the most units at location.

    Similar to the Command the Most, this condition compares
    the number of units at a specific location.
    The location can be restricted to certain elevations.
    """
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    return Condition(location, 0, 0, unit, 0, 7, 0, 0)


def MostKills(unit: Unit) -> Condition:  # noqa: N802
    """Current player has most kills of unit.

    This condition is considered true if the trigger's owner has
    the most kills of the specified Unit.
    """
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 8, 0, 0)


def HighestScore(ScoreType: _Score) -> Condition:  # noqa: N802, N803
    """Current player has highest score points.

    This condition is considered true if the trigger's owner has the highest
    Score. Note that if this is used as the only condition in a trigger,
    it will activate immediately at the start of the scenario,
    since all players will be tied for the highest score.
    """
    score_type = EncodeScore(ScoreType)
    return Condition(0, 0, 0, 0, 0, 9, score_type, 0)


def MostResources(resource_type: Resource) -> Condition:  # noqa: N802
    """Current player has most resources.

    Similar to Most Kills, this condition is considered true if the trigger's
    owner has the most of the specified resource.
    """
    resource_type = EncodeResource(resource_type)
    return Condition(0, 0, 0, 0, 0, 10, resource_type, 0)


def Switch(switch: _Switch, state: SwitchState) -> Condition:  # noqa: N802
    """Switch is set.

    This allows you to test against a switch value. Switches are on/off values
    that can be set with an action.
    Switches can be used to keep track of which triggers have been activated,
    to disable or enable certain triggers or to link multiple triggers
    together. You may also rename switches from this dialog box.
    """
    switch = EncodeSwitch(switch)
    state = EncodeSwitchState(state)
    return Condition(0, 0, 0, 0, state, 11, switch, 0)


def ElapsedTime(comparison: Comparison, time: Dword) -> Condition:  # noqa: N802
    """Elapsed scenario time is duration game seconds.

    This condition allows you to create triggers that occur after a specified
    number of game seconds have passed since the start of the scenario.
    """
    comparison = EncodeComparison(comparison)
    return Condition(0, 0, time, 0, comparison, 12, 0, 0)


def Opponents(  # noqa: N802
    player: Player, comparison: Comparison, number: Dword
) -> Condition:
    """Player has quantity opponents remaining in the game.

    This condition evaluates how many of the players are opponents of the
    trigger owner. By default, all of the other players are considered
    opponents. A player does not count as an opponent if either of the
    following conditions are met:

    * The player has been defeated. This condition only counts players
    that are still in the game.

    * The player is set for allied victory with the trigger owner,
    AND the player is set for allied victory with all other players set for
    allied victory with the trigger owner.
    (The enemy of an ally is still an enemy.)

    As a result, if opponents equals zero, all of remaining players are set for
    allied victory with each other. Use this condition with the Victory action
    to create a scenario that allows for allied victory.
    """
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    return Condition(0, player, number, 0, comparison, 14, 0, 0)


def Deaths(  # noqa: N802
    player: Player, comparison: Comparison, number: Dword, unit: Unit
) -> Condition:
    """Player has suffered quantity deaths of unit.

    Gives you the ability to create actions that are launched when a player has
    suffered a specific number of deaths of any of the units in the game.
    """
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(0, player, number, unit, comparison, 15, 0, 0)


def CommandLeast(unit: Unit) -> Condition:  # noqa: N802
    """Current player commands the least units.

    Command the Least allows you to define an action based on the player that
    commands the least units. You might use this to give advantages to slower
    players or to single out weakened players. Note that this condition checks
    all players, including neutral, computer controlled, and rescuable players.
    """
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 16, 0, 0)


def CommandLeastAt(unit: Unit, location: Location) -> Condition:  # noqa: N802
    """Current player commands the least units at location.

    Command the Least At is similar to 'Command the Least', however, but only
    compares units at a particular location.
    The location can be restricted to certain elevations.
    """
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    return Condition(location, 0, 0, unit, 0, 17, 0, 0)


def LeastKills(unit: Unit) -> Condition:  # noqa: N802
    """Current player has least kills of unit.

    This condition is considered true if the trigger's owner has the least
    kills of the specified Unit.
    """
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 18, 0, 0)


def LowestScore(ScoreType: _Score) -> Condition:  # noqa: N802, N803
    """Current player has lowest score points.

    This condition evaluates the current Score and is considered true
    if the current player has the lowest or is tied for the lowest score.
    Lowest Score checks all players, including neutral and rescuable players.
    """
    score_type = EncodeScore(ScoreType)
    return Condition(0, 0, 0, 0, 0, 19, score_type, 0)


def LeastResources(resource_type: Resource) -> Condition:  # noqa: N802
    """Current player has least resources.

    Similar to Least Kills, this condition is considered true if the trigger's
    owner has the least of the specified resource.
    Note that Least Resources checks all players, including neutral, computer
    controlled and rescuable players.
    """
    resource_type = EncodeResource(resource_type)
    return Condition(0, 0, 0, 0, 0, 20, resource_type, 0)


def Score(  # noqa: N802
    player: Player,
    ScoreType: _Score,  # noqa: N803
    comparison: Comparison,
    number: Dword,
) -> Condition:
    """Player score type score is quantity.

    This condition allows you to analyze a player's current Score and perform
    actions based on the value.
    You can reference any of the individual scoring types from score.
    """
    player = EncodePlayer(player)
    score_type = EncodeScore(ScoreType)
    comparison = EncodeComparison(comparison)
    return Condition(0, player, number, 0, comparison, 21, score_type, 0)


def Always() -> Condition:  # noqa: N802
    """Always.

    Accumulate requires that the player gather enough of a specific resource.
    """
    return Condition(0, 0, 0, 0, 0, 22, 0, 0)


def Never() -> Condition:  # noqa: N802
    """Never.

    The Never condition can be used to temporarily disable actions for testing.
    A trigger with the Never condition will not activate at any point.
    """
    return Condition(0, 0, 0, 0, 0, 23, 0, 0)


def Memory(dest: Dword, cmptype: Comparison, value: Dword) -> Condition:  # noqa: N802
    comparison = EncodeComparison(cmptype)
    return Deaths(EPD(dest), comparison, value, 0)


def MemoryEPD(dest: Dword, cmptype: Comparison, value: Dword) -> Condition:  # noqa: N802
    comparison = EncodeComparison(cmptype)
    return Deaths(dest, comparison, value, 0)


def DeathsX(  # noqa: N802
    player: Player,
    comparison: Comparison,
    number: Dword,
    unit: Unit,
    mask: Dword,
) -> Condition:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(mask, player, number, unit, comparison, 15, 0, 0, eudx=0x4353)


def MemoryX(  # noqa: N802
    dest: Dword, cmptype: Comparison, value: Dword, mask: Dword
) -> Condition:
    comparison = EncodeComparison(cmptype)
    return DeathsX(EPD(dest), comparison, value, 0, mask)


def MemoryXEPD(  # noqa: N802
    dest: Dword, cmptype: Comparison, value: Dword, mask: Dword
) -> Condition:
    comparison = EncodeComparison(cmptype)
    return DeathsX(dest, comparison, value, 0, mask)



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from ..core.eudstruct.vararray import _EUDVArray


class PVariable(_EUDVArray):
    def __init__(self, initvars=None, *, _from=None) -> None:
        super().__init__(initvars, _from=_from, _size=8, _basetype=None)

    def __getitem__(self, i):
        return super().__getitem__(c.EncodePlayer(i))

    def __setitem__(self, i, value) -> None:
        super().__setitem__(c.EncodePlayer(i), value)



from helper import *


@TestInstance
def test_ptr_read():
    a = EUDVariable(Db(b"\x01\x02\x03\x04\x05\x06\x07\x08"))

    # bread
    f_setcurpl(Player1)
    cp0 = f_getcurpl()
    br0 = f_bread(a + 0)
    cp1 = f_getcurpl()
    br1 = f_bread(a + 3)
    cp2 = f_getcurpl()
    test_equality("f_bread", [cp0, br0, cp1, br1, cp2], [0, 1, 0, 4, 0])

    # wread
    f_setcurpl(Player1)
    cp0 = f_getcurpl()
    wr0 = f_wread(a + 0)
    cp1 = f_getcurpl()
    wr1 = f_wread(a + 1)
    cp2 = f_getcurpl()
    wr2 = f_wread(a + 3)
    cp3 = f_getcurpl()
    test_equality(
        "f_wread",
        [cp0, wr0, cp1, wr1, cp2, wr2, cp3],
        [0, 0x0201, 0, 0x0302, 0, 0x0504, 0],
    )

    # dwread
    f_setcurpl(Player1)
    cp0 = f_getcurpl()
    dwr0 = f_dwread(a + 0)
    cp1 = f_getcurpl()
    dwr1 = f_dwread(a + 1)
    cp2 = f_getcurpl()
    dwr2 = f_dwread(a + 3)
    cp3 = f_getcurpl()
    test_equality(
        "f_dwread",
        [cp0, dwr0, cp1, dwr1, cp2, dwr2, cp3],
        [0, 0x04030201, 0, 0x05040302, 0, 0x07060504, 0],
    )


@TestInstance
def test_ptr_write():
    # bwrite
    f_setcurpl(Player1)
    a = Db(b"\x01\x02\x03\x04\x05\x06\x07\x08")
    cp0 = f_getcurpl()
    f_bwrite(a + 0, 5)
    cp1 = f_getcurpl()
    f_bwrite(a + 3, 7)
    cp2 = f_getcurpl()
    test_equality(
        "f_bwrite",
        [f_dwread_epd(EPD(a)), f_dwread_epd(EPD(a) + 1), cp0, cp1, cp2],
        [0x07030205, 0x08070605, 0, 0, 0],
    )

    # wwrite
    f_setcurpl(Player1)
    a = Db(b"\x01\x02\x03\x04\x05\x06\x07\x08")
    cp0 = f_getcurpl()
    f_wwrite(a + 0, 0xAABB)
    cp1 = f_getcurpl()
    f_wwrite(a + 3, 0xCCDD)
    cp2 = f_getcurpl()
    test_equality(
        "f_wwrite",
        [f_dwread_epd(EPD(a)), f_dwread_epd(EPD(a) + 1), cp0, cp1, cp2],
        [0xDD03AABB, 0x080706CC, 0, 0, 0],
    )


@TestInstance
def test_ptr_constmemio():
    DoActions(
        SetMemory(0x656EB0, SetTo, 0x04030201),
        SetMemory(0x656EB4, SetTo, 0x08070605),
    )
    a = 0x656EB0

    # bread
    f_setcurpl(Player1)
    cp0 = f_getcurpl()
    br0 = f_bread(a + 0)
    cp1 = f_getcurpl()
    br1 = f_bread(a + 3)
    cp2 = f_getcurpl()
    test_equality("f_bread(constexpr)", [cp0, br0, cp1, br1, cp2], [0, 1, 0, 4, 0])

    # wread
    f_setcurpl(Player1)
    cp0 = f_getcurpl()
    wr0 = f_wread(a + 0)
    cp1 = f_getcurpl()
    wr1 = f_wread(a + 1)
    cp2 = f_getcurpl()
    wr2 = f_wread(a + 3)
    cp3 = f_getcurpl()
    test_equality(
        "f_wread(constexpr)",
        [cp0, wr0, cp1, wr1, cp2, wr2, cp3],
        [0, 0x0201, 0, 0x0302, 0, 0x0504, 0],
    )

    # dwread
    f_setcurpl(Player1)
    cp0 = f_getcurpl()
    dwr0 = f_dwread(a + 0)
    cp1 = f_getcurpl()
    dwr1 = f_dwread(a + 1)
    cp2 = f_getcurpl()
    dwr2 = f_dwread(a + 3)
    cp3 = f_getcurpl()
    test_equality(
        "f_dwread(constexpr)",
        [cp0, dwr0, cp1, dwr1, cp2, dwr2, cp3],
        [0, 0x04030201, 0, 0x05040302, 0, 0x07060504, 0],
    )

    DoActions(
        SetMemory(0x656EB0, SetTo, 0x04030201),
        SetMemory(0x656EB4, SetTo, 0x08070605),
    )
    # bwrite
    f_setcurpl(Player1)
    a = 0x656EB0
    cp0 = f_getcurpl()
    f_bwrite(a + 0, 5)
    cp1 = f_getcurpl()
    f_bwrite(a + 3, 7)
    cp2 = f_getcurpl()
    test_equality(
        "f_bwrite(constexpr)",
        [f_dwread_epd(EPD(a)), f_dwread_epd(EPD(a) + 1), cp0, cp1, cp2],
        [0x07030205, 0x08070605, 0, 0, 0],
    )

    # wwrite
    f_setcurpl(Player1)
    DoActions(
        SetMemory(0x656EB0, SetTo, 0x04030201),
        SetMemory(0x656EB4, SetTo, 0x08070605),
    )
    a = 0x656EB0
    cp0 = f_getcurpl()
    f_wwrite(a + 0, 0xAABB)
    cp1 = f_getcurpl()
    f_wwrite(a + 3, 0xCCDD)
    cp2 = f_getcurpl()
    test_equality(
        "f_wwrite(constexpr)",
        [f_dwread_epd(EPD(a)), f_dwread_epd(EPD(a) + 1), cp0, cp1, cp2],
        [0xDD03AABB, 0x080706CC, 0, 0, 0],
    )



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...utils import EPD, _rand_lst, unProxy
from .. import allocator as ac
from .. import rawtrigger as rt
from .. import variable as ev
from ..eudfunc.eudf import EUDFunc, _EUDPredefineParam, _EUDPredefineReturn
from ..variable import SeqCompute
from ..variable.evcommon import _selfadder, _xv
from .muldiv import _quot

_x = _xv[0]


def f_bitand(a, b):
    """Calculate a & b"""
    return a & b


def f_bitor(a, b):
    """Calculate a | b"""
    return a | b


def f_bitxor(a, b):
    """Calculate a ^ b"""
    return a ^ b


def f_bitnand(a, b):
    """Calculate ~(a & b)"""
    try:
        return (a & b).iinvert()
    except AttributeError:
        return ~(a & b)


def f_bitnor(a, b):
    """Calculate ~(a | b)"""
    try:
        return (a | b).iinvert()
    except AttributeError:
        return ~(a | b)


def f_bitnxor(a, b):
    """Calculate ~(a ^ b)"""
    return f_bitxor(a, ~b)


def f_bitnot(a):
    """Calculate ~a"""
    return ~a


# -------


@EUDFunc
def f_bitsplit(a):
    """Splits bit of given number

    :returns: int bits[32];  // bits[i] = (ith bit from LSB of a is set)
    """
    bits = ev.EUDCreateVariables(32)
    rt.RawTrigger(actions=[bits[i].SetNumber(0) for i in range(32)])
    for i in range(31, -1, -1):
        rt.RawTrigger(conditions=a.AtLeastX(1, 2**i), actions=bits[i].SetNumber(1))
    return bits


# -------


@_EUDPredefineParam(1)
@_EUDPredefineReturn(1, 2)
@EUDFunc
def _exp2_vv(n):
    ret = _exp2_vv._frets[0]
    ret << 0
    for i in _rand_lst(range(32)):
        rt.RawTrigger(conditions=[n == i], actions=ret.SetNumber(2**i))
    # return ret


def _exp2(n):
    if isinstance(n, int):
        if n >= 32:
            return 0
        return 1 << n

    return _exp2_vv(n)


@EUDFunc
def _f_bitrshift(a, b):
    ret = ev.EUDVariable()

    skip_check = ac.Forward()
    exponential = ac.Forward()
    skip_to_here = ac.Forward()
    return_function = ac.Forward()

    skip_check << rt.RawTrigger(
        nextptr=exponential,
        conditions=b.AtLeast(32),
        actions=[ret.SetNumber(0), rt.SetNextPtr(skip_check, skip_to_here)],
    )
    skip_to_here << rt.RawTrigger(
        nextptr=return_function, actions=rt.SetNextPtr(skip_check, exponential)
    )
    exponential << rt.NextTrigger()
    ret << _quot(a, _exp2_vv(b))
    return_function << rt.NextTrigger()
    return ret


@EUDFunc
def _f_bitlshift(a, b):
    loopstart = ac.Forward()
    loopend = ac.Forward()
    loopcnt = ac.Forward()

    rt.RawTrigger(actions=[rt.SetNextPtr(a.GetVTable(), loopcnt), *a.QueueAddTo(a)])

    loopstart << rt.RawTrigger(
        nextptr=loopend,
        conditions=[a.AtLeast(1), b.AtLeast(1)],
        actions=rt.SetNextPtr(loopstart, a.GetVTable()),
    )
    loopcnt << rt.RawTrigger(
        nextptr=loopstart,
        actions=[b.SubtractNumber(1), rt.SetNextPtr(loopstart, loopend)],
    )
    loopend << rt.NextTrigger()

    return a


def f_bitlshift(a, b, _fdict={}, **kwargs):
    """Calculate a << b"""
    a, b = unProxy(a), unProxy(b)
    if not isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if "ret" in kwargs:
            SeqCompute((kwargs["ret"][0], rt.SetTo, a << b))
            return kwargs["ret"][0]
        return a << b
    ret = kwargs["ret"][0] if "ret" in kwargs else ev.EUDVariable()
    if isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if b == 0:
            if a is ret:
                return a
            ev.SeqCompute([(ret, rt.SetTo, a)])
        elif 1 <= b <= 10:
            try:
                f = _fdict[b]
            except KeyError:
                set_ret = _selfadder.SetDest(0)

                @_EUDPredefineParam((EPD(set_ret) + 5, _selfadder))
                @EUDFunc
                def f(ret, adder):
                    for i in range(b):
                        ev.VProc(adder, [])
                    ev.VProc(
                        adder,
                        [
                            set_ret,
                            adder.SetModifier(rt.SetTo),
                        ],
                    )
                    rt.RawTrigger(
                        actions=[
                            adder.SetDest(EPD(adder.getValueAddr())),
                            adder.SetModifier(rt.Add),
                        ]
                    )

                _fdict[b] = f

            if ev.IsEUDVariable(ret):
                retarg = EPD(ret.getValueAddr())
            else:
                retarg = ret
            if "ret" in kwargs:
                del kwargs["ret"]
            f(retarg, a, **kwargs)
        else:
            if a is not ret:
                ev.SeqCompute([(ret, rt.SetTo, a)])
            ev.vbase.VariableBase.__ilshift__(ret, b)
        return ret

    return _f_bitlshift(a, b, **kwargs)


def f_bitrshift(a, b, **kwargs):
    """Calculate a >> b"""
    a, b = unProxy(a), unProxy(b)
    if not isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if "ret" in kwargs:
            SeqCompute((kwargs["ret"][0], rt.SetTo, a >> b))
            return kwargs["ret"][0]
        return a >> b
    ret = kwargs["ret"][0] if "ret" in kwargs else ev.EUDVariable()
    if isinstance(a, ev.EUDVariable) and not isinstance(b, ev.EUDVariable):
        if b == 0:
            if a is ret:
                return a
            ev.SeqCompute([(ret, rt.SetTo, a)])
        else:
            if a is not ret:
                ev.SeqCompute([(ret, rt.SetTo, a)])
            ev.vbase.VariableBase.__irshift__(ret, b)
        return ret

    return _f_bitrshift(a, b, **kwargs)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import random
import warnings

from .. import core as c
from .. import ctrlstru as cs
from .. import utils as ut
from ..core.eudfunc.eudf import _EUDPredefineParam, _EUDPredefineReturn
from ..localize import _
from . import modcurpl as cp
from . import readtable


@_EUDPredefineReturn(2)
@_EUDPredefineParam(cp.CP)
@c.EUDFunc
def f_dwepdread_epd(targetplayer):
    ptr, epd = f_dwepdread_epd._frets
    u = random.randint(234, 65535)
    acts = [
        ptr.SetNumber(0),
        epd.SetNumber(ut.EPD(0)),
        c.SetMemory(0x6509B0, c.Add, -12 * u),
    ]
    cs.DoActions(ut._rand_lst(acts))

    for i in ut._rand_lst(range(32)):
        acts = [
            ptr.AddNumber(2**i),
            epd.AddNumber(2 ** (i - 2)) if i >= 2 else [],
        ]
        c.RawTrigger(
            conditions=c.DeathsX(cp.CP, c.AtLeast, 1, u, 2**i),
            actions=ut._rand_lst(acts),
        )

    cp.f_setcurpl2cpcache()

    # return ptr, epd


def f_dwread_epd(targetplayer, *, ret=None):
    return readtable._epd_caller(readtable._get(0xFFFFFFFF, 0))(
        targetplayer, ret=ret
    )


def f_epdread_epd(targetplayer, *, ret=None):
    return readtable._epd_caller(
        readtable._insert_or_get(0xFFFFFFFC, -2, True),
        _operations=[(ut.EPD(readtable.copy_ret) + 5, c.SetTo, ut.EPD(0))],
    )(targetplayer, ret=ret)


# Special flag reading functions
def f_flagread_epd(targetplayer, *flags, _readerdict={}):
    flags = tuple(flags)  # Make flags hashable

    if flags in _readerdict:
        readerf = _readerdict[flags]
    else:
        # Create reader function
        @_EUDPredefineParam(cp.CP)
        @c.EUDFunc
        def readerf(targetplayer):
            flagsv = c.EUDCreateVariables(len(flags))

            # All set to 0
            cs.DoActions([flagv.SetNumber(0) for flagv in flagsv])

            # Fill flags
            for i in range(31, -1, -1):
                bitandflags = [flag & (2**i) for flag in flags]
                if sum(bitandflags) == 0:
                    continue
                c.RawTrigger(
                    conditions=[c.DeathsX(cp.CP, c.AtLeast, 1, 0, 2**i)],
                    actions=[
                        flagv.AddNumber(2**i)
                        for j, flagv in enumerate(flagsv)
                        if flags[j] & (2**i)
                    ],
                )

            cp.f_setcurpl2cpcache()

            return flagsv

        _readerdict[flags] = readerf

    return readerf(targetplayer)


# Writing functions


def setdw_epd(targetplayer, modifier, value):
    if c.IsEUDVariable(value):
        if c.IsEUDVariable(targetplayer):
            if targetplayer is value:
                c.VProc(
                    value,
                    [
                        value.SetDest(value),
                        value.SetModifier(c.SetTo),
                    ],
                )
                return c.VProc(value, value.SetModifier(modifier))
            return c.VProc(
                [targetplayer, value],
                [
                    targetplayer.QueueAssignTo(ut.EPD(value.getDestAddr())),
                    value.SetModifier(modifier),
                ],
            )
        return c.VProc(
            value,
            [
                value.SetDest(targetplayer),
                value.SetModifier(modifier),
            ],
        )
    cs.DoActions(c.SetDeaths(targetplayer, modifier, value, 0))


def f_dwwrite_epd(targetplayer, value):
    setdw_epd(targetplayer, c.SetTo, value)


def f_dwadd_epd(targetplayer, value):
    setdw_epd(targetplayer, c.Add, value)


def f_dwsubtract_epd(targetplayer, value):
    setdw_epd(targetplayer, c.Subtract, value)


# Dword breaking functions
@c.EUDFunc
def f_dwbreak(number):
    """Get hiword/loword/4byte of dword"""
    word = [c.EUDXVariable(0, c.SetTo, 0, 0xFFFF), c.EUDVariable()]
    byte = [c.EUDXVariable(0, c.SetTo, 0, 0xFF), *c.EUDCreateVariables(3)]

    # Clear byte[], word[]
    c.VProc(
        [number, word[0]],
        [
            number.SetDest(word[0]),
            word[0].SetDest(byte[0]),
            word[1].SetNumber(0),
            byte[1].SetNumber(0),
            byte[2].SetNumber(0),
            byte[3].SetNumber(0),
        ],
    )

    for i in ut._rand_lst(range(8, 32)):
        byteidx = i // 8
        wordidx = i // 16
        byteexp = i % 8
        wordexp = i % 16

        c.RawTrigger(
            conditions=number.AtLeastX(1, 2**i),
            actions=[
                byte[byteidx].AddNumber(2**byteexp),
                word[wordidx].AddNumber(2**wordexp) if wordidx == 1 else [],
            ],
        )

    return word[0], word[1], byte[0], byte[1], byte[2], byte[3]


@c.EUDFunc
def f_dwbreak2(number):
    """Get hiword/loword/4byte of dword"""
    word = c.EUDCreateVariables(2)
    byte = c.EUDCreateVariables(4)

    # Set byte[], word[]
    c.VProc(
        [number, word[0], word[1], byte[0], byte[1], byte[2]],
        [
            number.SetDest(word[0]),
            word[0].SetDest(word[1]),
            word[1].SetDest(byte[0]),
            byte[0].SetDest(byte[1]),
            byte[1].SetDest(byte[2]),
            byte[2].SetDest(byte[3]),
        ],
    )
    c.RawTrigger(
        actions=[
            word[0].SetNumberX(0, ~0xFFFF),
            word[1].SetNumberX(0, 0xFFFF),
            byte[0].SetNumberX(0, ~0xFF),
            byte[1].SetNumberX(0, ~0xFF00),
            byte[2].SetNumberX(0, ~0xFF0000),
            byte[3].SetNumberX(0, 0xFFFFFF),
        ]
    )

    return word[0], word[1], byte[0], byte[1], byte[2], byte[3]


# backward compatibility functions
E = _("safe read functions are deprecated in 0.61 and will be removed in 0.63")


def f_dwepdread_epd_safe(*args, **kwargs):
    warnings.warn(E, DeprecationWarning)
    return f_dwepdread_epd(*args, **kwargs)


def f_dwread_epd_safe(*args, **kwargs):
    warnings.warn(E, DeprecationWarning)
    return f_dwread_epd(*args, **kwargs)


def f_epdread_epd_safe(*args, **kwargs):
    warnings.warn(E, DeprecationWarning)
    return f_epdread_epd(*args, **kwargs)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from __future__ import annotations

import functools
from collections.abc import Iterator
from typing import TYPE_CHECKING

from .. import core as c
from ..collections import EUDArray
from ..core import EUDVariable
from ..ctrlstru import EUDEndIf, EUDIfNot, EUDJumpIf

if TYPE_CHECKING:
    from ..core.rawtrigger.typehint import Player

_runner_start: list[c.Forward] = [c.Forward() for _ in range(8)]
_runner_end: list[c.Forward] = [c.Forward() for _ in range(8)]

_runner_cp: c.EUDLightVariable = c.EUDLightVariable()

c.PushTriggerScope()
for player in range(8):
    _runner_start[player] << c.RawTrigger(nextptr=_runner_end[player])
    _runner_end[player] << c.RawTrigger(nextptr=~(0x51A280 + player * 12 + 4))
c.PopTriggerScope()


@c.EUDFunc
def RunTrigTrigger() -> None:  # noqa: N802
    from ..memio import f_getcurpl, f_setcurpl

    oldcp = f_getcurpl()

    for player in range(8):
        skipt = c.Forward()
        EUDJumpIf(
            c.Memory(
                0x51A280 + 12 * player + 4,
                c.Exactly,
                0x51A280 + 12 * player + 4,
            ),
            skipt,
        )
        nt = c.Forward()
        c.RawTrigger(
            nextptr=_runner_start[player],
            actions=[
                c.SetMemory(0x6509B0, c.SetTo, player),
                _runner_cp.SetNumber(player),
                c.SetNextPtr(_runner_end[player], nt),
            ],
        )
        nt << c.RawTrigger(
            actions=c.SetNextPtr(_runner_end[player], ~(0x51A280 + player * 12 + 4))
        )
        skipt << c.NextTrigger()

    f_setcurpl(oldcp)


#######


@functools.cache
def _alloc_trigtrigger_link() -> tuple[EUDArray, EUDArray, EUDArray]:
    return EUDArray(8), EUDArray(8), EUDArray(_runner_end)


def GetFirstTrigTrigger(player: Player) -> EUDVariable:  # noqa: N802
    """Get dlist start of trig-trigger for player"""
    player = c.EncodePlayer(player)
    orig_tstart, _orig_tend, _runner_end_array = _alloc_trigtrigger_link()
    return orig_tstart[player]


def GetLastTrigTrigger(player: Player) -> EUDVariable:  # noqa: N802
    """Get dlist end of trig-trigger for player"""
    player = c.EncodePlayer(player)
    _orig_tstart, orig_tend, _runner_end_array = _alloc_trigtrigger_link()
    return orig_tend[player]


def TrigTriggerBegin(player: Player) -> EUDVariable:  # noqa: N802
    return GetFirstTrigTrigger(player)


def TrigTriggerEnd(player: Player) -> c.Forward | EUDVariable:  # noqa: N802
    _orig_tstart, _orig_tend, runner_end_array = _alloc_trigtrigger_link()
    player = c.EncodePlayer(player)
    if isinstance(player, int):
        return _runner_end[player]
    return runner_end_array[player]


#######


def EUDLoopTrigger(player) -> Iterator[tuple[EUDVariable, EUDVariable]]:  # noqa: N802
    from ..eudlib.utilf.listloop import EUDLoopList

    player = c.EncodePlayer(player)

    tbegin = TrigTriggerBegin(player)
    if EUDIfNot()(tbegin == 0):
        tend = TrigTriggerEnd(player)
        yield from EUDLoopList(tbegin, tend)
    EUDEndIf()



from __future__ import annotations

import pytest

import env  # noqa: F401
from pybind11_tests import ConstructorStats
from pybind11_tests import multiple_inheritance as m


def test_multiple_inheritance_cpp():
    mt = m.MIType(3, 4)

    assert mt.foo() == 3
    assert mt.bar() == 4


@pytest.mark.xfail("env.PYPY")
def test_multiple_inheritance_mix1():
    class Base1:
        def __init__(self, i):
            self.i = i

        def foo(self):
            return self.i

    class MITypePy(Base1, m.Base2):
        def __init__(self, i, j):
            Base1.__init__(self, i)
            m.Base2.__init__(self, j)

    mt = MITypePy(3, 4)

    assert mt.foo() == 3
    assert mt.bar() == 4


def test_multiple_inheritance_mix2():
    class Base2:
        def __init__(self, i):
            self.i = i

        def bar(self):
            return self.i

    class MITypePy(m.Base1, Base2):
        def __init__(self, i, j):
            m.Base1.__init__(self, i)
            Base2.__init__(self, j)

    mt = MITypePy(3, 4)

    assert mt.foo() == 3
    assert mt.bar() == 4


@pytest.mark.xfail("env.PYPY")
def test_multiple_inheritance_python():
    class MI1(m.Base1, m.Base2):
        def __init__(self, i, j):
            m.Base1.__init__(self, i)
            m.Base2.__init__(self, j)

    class B1:
        def v(self):
            return 1

    class MI2(B1, m.Base1, m.Base2):
        def __init__(self, i, j):
            B1.__init__(self)
            m.Base1.__init__(self, i)
            m.Base2.__init__(self, j)

    class MI3(MI2):
        def __init__(self, i, j):
            MI2.__init__(self, i, j)

    class MI4(MI3, m.Base2):
        def __init__(self, i, j):
            MI3.__init__(self, i, j)
            # This should be ignored (Base2 is already initialized via MI2):
            m.Base2.__init__(self, i + 100)

    class MI5(m.Base2, B1, m.Base1):
        def __init__(self, i, j):
            B1.__init__(self)
            m.Base1.__init__(self, i)
            m.Base2.__init__(self, j)

    class MI6(m.Base2, B1):
        def __init__(self, i):
            m.Base2.__init__(self, i)
            B1.__init__(self)

    class B2(B1):
        def v(self):
            return 2

    class B3:
        def v(self):
            return 3

    class B4(B3, B2):
        def v(self):
            return 4

    class MI7(B4, MI6):
        def __init__(self, i):
            B4.__init__(self)
            MI6.__init__(self, i)

    class MI8(MI6, B3):
        def __init__(self, i):
            MI6.__init__(self, i)
            B3.__init__(self)

    class MI8b(B3, MI6):
        def __init__(self, i):
            B3.__init__(self)
            MI6.__init__(self, i)

    mi1 = MI1(1, 2)
    assert mi1.foo() == 1
    assert mi1.bar() == 2

    mi2 = MI2(3, 4)
    assert mi2.v() == 1
    assert mi2.foo() == 3
    assert mi2.bar() == 4

    mi3 = MI3(5, 6)
    assert mi3.v() == 1
    assert mi3.foo() == 5
    assert mi3.bar() == 6

    mi4 = MI4(7, 8)
    assert mi4.v() == 1
    assert mi4.foo() == 7
    assert mi4.bar() == 8

    mi5 = MI5(10, 11)
    assert mi5.v() == 1
    assert mi5.foo() == 10
    assert mi5.bar() == 11

    mi6 = MI6(12)
    assert mi6.v() == 1
    assert mi6.bar() == 12

    mi7 = MI7(13)
    assert mi7.v() == 4
    assert mi7.bar() == 13

    mi8 = MI8(14)
    assert mi8.v() == 1
    assert mi8.bar() == 14

    mi8b = MI8b(15)
    assert mi8b.v() == 3
    assert mi8b.bar() == 15


def test_multiple_inheritance_python_many_bases():
    class MIMany14(m.BaseN1, m.BaseN2, m.BaseN3, m.BaseN4):
        def __init__(self):
            m.BaseN1.__init__(self, 1)
            m.BaseN2.__init__(self, 2)
            m.BaseN3.__init__(self, 3)
            m.BaseN4.__init__(self, 4)

    class MIMany58(m.BaseN5, m.BaseN6, m.BaseN7, m.BaseN8):
        def __init__(self):
            m.BaseN5.__init__(self, 5)
            m.BaseN6.__init__(self, 6)
            m.BaseN7.__init__(self, 7)
            m.BaseN8.__init__(self, 8)

    class MIMany916(
        m.BaseN9,
        m.BaseN10,
        m.BaseN11,
        m.BaseN12,
        m.BaseN13,
        m.BaseN14,
        m.BaseN15,
        m.BaseN16,
    ):
        def __init__(self):
            m.BaseN9.__init__(self, 9)
            m.BaseN10.__init__(self, 10)
            m.BaseN11.__init__(self, 11)
            m.BaseN12.__init__(self, 12)
            m.BaseN13.__init__(self, 13)
            m.BaseN14.__init__(self, 14)
            m.BaseN15.__init__(self, 15)
            m.BaseN16.__init__(self, 16)

    class MIMany19(MIMany14, MIMany58, m.BaseN9):
        def __init__(self):
            MIMany14.__init__(self)
            MIMany58.__init__(self)
            m.BaseN9.__init__(self, 9)

    class MIMany117(MIMany14, MIMany58, MIMany916, m.BaseN17):
        def __init__(self):
            MIMany14.__init__(self)
            MIMany58.__init__(self)
            MIMany916.__init__(self)
            m.BaseN17.__init__(self, 17)

    # Inherits from 4 registered C++ classes: can fit in one pointer on any modern arch:
    a = MIMany14()
    for i in range(1, 4):
        assert getattr(a, "f" + str(i))() == 2 * i

    # Inherits from 8: requires 1/2 pointers worth of holder flags on 32/64-bit arch:
    b = MIMany916()
    for i in range(9, 16):
        assert getattr(b, "f" + str(i))() == 2 * i

    # Inherits from 9: requires >= 2 pointers worth of holder flags
    c = MIMany19()
    for i in range(1, 9):
        assert getattr(c, "f" + str(i))() == 2 * i

    # Inherits from 17: requires >= 3 pointers worth of holder flags
    d = MIMany117()
    for i in range(1, 17):
        assert getattr(d, "f" + str(i))() == 2 * i


def test_multiple_inheritance_virtbase():
    class MITypePy(m.Base12a):
        def __init__(self, i, j):
            m.Base12a.__init__(self, i, j)

    mt = MITypePy(3, 4)
    assert mt.bar() == 4
    assert m.bar_base2a(mt) == 4
    assert m.bar_base2a_sharedptr(mt) == 4


def test_mi_static_properties():
    """Mixing bases with and without static properties should be possible
    and the result should be independent of base definition order"""

    for d in (m.VanillaStaticMix1(), m.VanillaStaticMix2()):
        assert d.vanilla() == "Vanilla"
        assert d.static_func1() == "WithStatic1"
        assert d.static_func2() == "WithStatic2"
        assert d.static_func() == d.__class__.__name__

        m.WithStatic1.static_value1 = 1
        m.WithStatic2.static_value2 = 2
        assert d.static_value1 == 1
        assert d.static_value2 == 2
        assert d.static_value == 12

        d.static_value1 = 0
        assert d.static_value1 == 0
        d.static_value2 = 0
        assert d.static_value2 == 0
        d.static_value = 0
        assert d.static_value == 0


# Requires PyPy 6+
def test_mi_dynamic_attributes():
    """Mixing bases with and without dynamic attribute support"""

    for d in (m.VanillaDictMix1(), m.VanillaDictMix2()):
        d.dynamic = 1
        assert d.dynamic == 1


def test_mi_unaligned_base():
    """Returning an offset (non-first MI) base class pointer should recognize the instance"""

    n_inst = ConstructorStats.detail_reg_inst()

    c = m.I801C()
    d = m.I801D()
    # + 4 below because we have the two instances, and each instance has offset base I801B2
    assert ConstructorStats.detail_reg_inst() == n_inst + 4
    b1c = m.i801b1_c(c)
    assert b1c is c
    b2c = m.i801b2_c(c)
    assert b2c is c
    b1d = m.i801b1_d(d)
    assert b1d is d
    b2d = m.i801b2_d(d)
    assert b2d is d

    assert ConstructorStats.detail_reg_inst() == n_inst + 4  # no extra instances
    del c, b1c, b2c
    assert ConstructorStats.detail_reg_inst() == n_inst + 2
    del d, b1d, b2d
    assert ConstructorStats.detail_reg_inst() == n_inst


def test_mi_base_return():
    """Tests returning an offset (non-first MI) base class pointer to a derived instance"""

    n_inst = ConstructorStats.detail_reg_inst()

    c1 = m.i801c_b1()
    assert type(c1) is m.I801C
    assert c1.a == 1
    assert c1.b == 2

    d1 = m.i801d_b1()
    assert type(d1) is m.I801D
    assert d1.a == 1
    assert d1.b == 2

    assert ConstructorStats.detail_reg_inst() == n_inst + 4

    c2 = m.i801c_b2()
    assert type(c2) is m.I801C
    assert c2.a == 1
    assert c2.b == 2

    d2 = m.i801d_b2()
    assert type(d2) is m.I801D
    assert d2.a == 1
    assert d2.b == 2

    assert ConstructorStats.detail_reg_inst() == n_inst + 8

    del c2
    assert ConstructorStats.detail_reg_inst() == n_inst + 6
    del c1, d1, d2
    assert ConstructorStats.detail_reg_inst() == n_inst

    # Returning an unregistered derived type with a registered base; we won't
    # pick up the derived type, obviously, but should still work (as an object
    # of whatever type was returned).
    e1 = m.i801e_c()
    assert type(e1) is m.I801C
    assert e1.a == 1
    assert e1.b == 2

    e2 = m.i801e_b2()
    assert type(e2) is m.I801B2
    assert e2.b == 2


def test_diamond_inheritance():
    """Tests that diamond inheritance works as expected (issue #959)"""

    # Issue #959: this shouldn't segfault:
    d = m.D()

    # Make sure all the various distinct pointers are all recognized as registered instances:
    assert d is d.c0()
    assert d is d.c1()
    assert d is d.b()
    assert d is d.c0().b()
    assert d is d.c1().b()
    assert d is d.c0().c1().b().c0().b()


def test_pr3635_diamond_b():
    o = m.MVB()
    assert o.b == 1

    assert o.get_b_b() == 1


def test_pr3635_diamond_c():
    o = m.MVC()
    assert o.b == 1
    assert o.c == 2

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1

    assert o.get_c_c() == 2


def test_pr3635_diamond_d0():
    o = m.MVD0()
    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d0_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d0_c() == 2

    assert o.get_d0_d0() == 3


def test_pr3635_diamond_d1():
    o = m.MVD1()
    assert o.b == 1
    assert o.c == 2
    assert o.d1 == 4

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d1_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d1_c() == 2

    assert o.get_d1_d1() == 4


def test_pr3635_diamond_e():
    o = m.MVE()
    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3
    assert o.d1 == 4
    assert o.e == 5

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d0_b() == 1
    assert o.get_d1_b() == 1
    assert o.get_e_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d0_c() == 2
    assert o.get_d1_c() == 2
    assert o.get_e_c() == 2

    assert o.get_d0_d0() == 3
    assert o.get_e_d0() == 3

    assert o.get_d1_d1() == 4
    assert o.get_e_d1() == 4

    assert o.get_e_e() == 5


def test_pr3635_diamond_f():
    o = m.MVF()
    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3
    assert o.d1 == 4
    assert o.e == 5
    assert o.f == 6

    assert o.get_b_b() == 1
    assert o.get_c_b() == 1
    assert o.get_d0_b() == 1
    assert o.get_d1_b() == 1
    assert o.get_e_b() == 1
    assert o.get_f_b() == 1

    assert o.get_c_c() == 2
    assert o.get_d0_c() == 2
    assert o.get_d1_c() == 2
    assert o.get_e_c() == 2
    assert o.get_f_c() == 2

    assert o.get_d0_d0() == 3
    assert o.get_e_d0() == 3
    assert o.get_f_d0() == 3

    assert o.get_d1_d1() == 4
    assert o.get_e_d1() == 4
    assert o.get_f_d1() == 4

    assert o.get_e_e() == 5
    assert o.get_f_e() == 5

    assert o.get_f_f() == 6


def test_python_inherit_from_mi():
    """Tests extending a Python class from a single inheritor of a MI class"""

    class PyMVF(m.MVF):
        g = 7

        def get_g_g(self):
            return self.g

    o = PyMVF()

    assert o.b == 1
    assert o.c == 2
    assert o.d0 == 3
    assert o.d1 == 4
    assert o.e == 5
    assert o.f == 6
    assert o.g == 7

    assert o.get_g_g() == 7



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from collections.abc import Callable
from typing import NoReturn

from ..localize import _


class CtrlStruOpener:
    def __init__(self, f: Callable) -> None:
        self._f: Callable = f
        self._called = False

    def __del__(self) -> None:
        if not self._called:
            raise RuntimeError(
                _("Control structures must be double-parenthesised.")
                + "\n ex) EUDInfLoop()()"
            )

    def __bool__(self) -> NoReturn:
        raise RuntimeError(
            _("Control structures must be double-parenthesised.")
            + "\n ex) EUDInfLoop()()"
        )

    def __call__(self, *args, **kwargs):
        self._called = True
        return self._f(*args, **kwargs)



from helper import *


@TestInstance
def test_dwpatch():
    from eudplib.eudlib.utilf.mempatch import (
        f_dwpatch_epd,
        f_blockpatch_epd,
        f_unpatchall,
    )

    # Initial value
    DoActions(SetMemory(0x58A364, SetTo, 1234), SetMemory(0x58A368, SetTo, 5678))

    # f_dwpatch
    f_dwpatch_epd(EPD(0x58A364), 123)
    f_dwpatch_epd(EPD(0x58A368), 456)
    test_assert(
        "f_dwpatch test",
        [f_dwread_epd(EPD(0x58A364)) == 123, f_dwread_epd(EPD(0x58A368)) == 456],
    )

    # f_blockpatch
    data = Db(b"\x01\x02\x03\x04\x05\x06\x07\x08")
    f_blockpatch_epd(EPD(0x58A364), EPD(data), 2)
    test_assert(
        "f_blockpatch_epd test",
        [
            f_dwread_epd(EPD(0x58A364)) == 0x04030201,
            f_dwread_epd(EPD(0x58A368)) == 0x08070605,
        ],
    )

    # unpatch
    f_unpatchall()
    test_assert(
        "f_unpatchall test (Relies on patches above)",
        [f_dwread_epd(EPD(0x58A364)) == 1234, f_dwread_epd(EPD(0x58A368)) == 5678],
    )

    # reset
    DoActions(SetMemory(0x58A364, SetTo, 0), SetMemory(0x58A368, SetTo, 0))



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from struct import pack
from typing import TypeAlias

from ..core import UnitProperty

# re-export stuffs
from ..core.rawtrigger.constenc import Add as Add
from ..core.rawtrigger.constenc import All as All
from ..core.rawtrigger.constenc import AlliedVictory as AlliedVictory
from ..core.rawtrigger.constenc import Ally as Ally
from ..core.rawtrigger.constenc import AtLeast as AtLeast
from ..core.rawtrigger.constenc import AtMost as AtMost
from ..core.rawtrigger.constenc import Attack as Attack
from ..core.rawtrigger.constenc import Buildings as Buildings
from ..core.rawtrigger.constenc import Clear as Clear
from ..core.rawtrigger.constenc import Cleared as Cleared
from ..core.rawtrigger.constenc import Custom as Custom
from ..core.rawtrigger.constenc import Disable as Disable
from ..core.rawtrigger.constenc import Enable as Enable
from ..core.rawtrigger.constenc import (
    EncodeAllyStatus,
    EncodeComparison,
    EncodeCount,
    EncodeModifier,
    EncodeOrder,
    EncodePlayer,
    EncodeProperty,
    EncodePropState,
    EncodeResource,
    EncodeScore,
    EncodeSwitchAction,
    EncodeSwitchState,
    TrgAllyStatus,
    TrgComparison,
    TrgCount,
    TrgModifier,
    TrgOrder,
    TrgPropState,
    TrgResource,
    TrgScore,
    TrgSwitchAction,
    TrgSwitchState,
    _Kills,
)
from ..core.rawtrigger.constenc import Enemy as Enemy
from ..core.rawtrigger.constenc import Exactly as Exactly
from ..core.rawtrigger.constenc import Gas as Gas
from ..core.rawtrigger.constenc import KillsAndRazings as KillsAndRazings
from ..core.rawtrigger.constenc import Move as Move
from ..core.rawtrigger.constenc import Ore as Ore
from ..core.rawtrigger.constenc import OreAndGas as OreAndGas
from ..core.rawtrigger.constenc import Patrol as Patrol
from ..core.rawtrigger.constenc import Random as Random
from ..core.rawtrigger.constenc import Razings as Razings
from ..core.rawtrigger.constenc import Set as Set
from ..core.rawtrigger.constenc import SetTo as SetTo
from ..core.rawtrigger.constenc import Subtract as Subtract
from ..core.rawtrigger.constenc import Toggle as Toggle
from ..core.rawtrigger.constenc import Total as Total
from ..core.rawtrigger.constenc import Units as Units
from ..core.rawtrigger.constenc import UnitsAndBuildings as UnitsAndBuildings
from ..core.rawtrigger.strenc import (
    EncodeAIScript,
    EncodeLocation,
    EncodeString,
    EncodeSwitch,
    EncodeUnit,
)
from ..scdata.player import P1 as P1
from ..scdata.player import P2 as P2
from ..scdata.player import P3 as P3
from ..scdata.player import P4 as P4
from ..scdata.player import P5 as P5
from ..scdata.player import P6 as P6
from ..scdata.player import P7 as P7
from ..scdata.player import P8 as P8
from ..scdata.player import P9 as P9
from ..scdata.player import P10 as P10
from ..scdata.player import P11 as P11
from ..scdata.player import P12 as P12
from ..scdata.player import Allies as Allies
from ..scdata.player import AllPlayers as AllPlayers
from ..scdata.player import CurrentPlayer as CurrentPlayer
from ..scdata.player import Foes as Foes
from ..scdata.player import Force1 as Force1
from ..scdata.player import Force2 as Force2
from ..scdata.player import Force3 as Force3
from ..scdata.player import Force4 as Force4
from ..scdata.player import NeutralPlayers as NeutralPlayers
from ..scdata.player import (
    NonAlliedVictoryPlayers as NonAlliedVictoryPlayers,
)
from ..scdata.player import Player1 as Player1
from ..scdata.player import Player2 as Player2
from ..scdata.player import Player3 as Player3
from ..scdata.player import Player4 as Player4
from ..scdata.player import Player5 as Player5
from ..scdata.player import Player6 as Player6
from ..scdata.player import Player7 as Player7
from ..scdata.player import Player8 as Player8
from ..scdata.player import Player9 as Player9
from ..scdata.player import Player10 as Player10
from ..scdata.player import Player11 as Player11
from ..scdata.player import Player12 as Player12
from ..scdata.player import TrgPlayer
from ..utils import EPD, FlattenList, ep_assert

"""
Defines stock conditions & actions. You are most likely to use only conditions
declared in this file. This file also serves as a basic reference for trigger
condition / actions.

 ex) You want to create 'Create Unit' action:
   1. Remove spaces. 'CreateUnit'
   2. Find 'CreateUnit' in this file. You'll see the following function def.
     def CreateUnit(Number, unit, where, for_player):
   3. Make your action as mentioned here.
     ex) CreateUnit(30, 'Terran Marine', 'Anywhere', Player1)
"""
AllyStatus: TypeAlias = TrgAllyStatus | int
Comparison: TypeAlias = TrgComparison | int
Count: TypeAlias = TrgCount | int
Modifier: TypeAlias = TrgModifier | int
_Order: TypeAlias = TrgOrder | int
Player: TypeAlias = TrgPlayer | int
PropState: TypeAlias = TrgPropState | int
Resource: TypeAlias = TrgResource | int
_Score: TypeAlias = TrgScore | int
SwitchAction: TypeAlias = TrgSwitchAction | int
SwitchState: TypeAlias = TrgSwitchState | int

Unit: TypeAlias = str | int | bytes
Location: TypeAlias = str | int | bytes
String: TypeAlias = str | int | bytes
AIScript: TypeAlias = str | int | bytes
_Switch: TypeAlias = str | int | bytes


def Condition(  # noqa: N802
    locid: int,
    player: int,
    amount: int,
    unitid: int,
    comparison: int,
    condtype: int,
    restype: int,
    flag: int,
    *,
    eudx: int = 0,
) -> bytes:
    if player < 0:
        player += 0x100000000  # EPD

    player &= 0xFFFFFFFF
    amount &= 0xFFFFFFFF

    return pack(
        "<IIIHBBBBH",
        locid,
        player,
        amount,
        unitid,
        comparison,
        condtype,
        restype,
        flag,
        eudx,
    )


def Action(  # noqa: N802
    locid1: int,
    strid: int,
    wavid: int,
    time: int,
    player1: int,
    player2: int,
    unitid: int,
    acttype: int,
    amount: int,
    flags: int,
    *,
    eudx: int = 0,
) -> bytes:
    if player1 < 0:
        player1 += 0x100000000  # EPD
    if player2 < 0:
        player2 += 0x100000000  # EPD

    player1 &= 0xFFFFFFFF
    player2 &= 0xFFFFFFFF

    return pack(
        "<IIIIIIHBBBBH",
        locid1,
        strid,
        wavid,
        time,
        player1,
        player2,
        unitid,
        acttype,
        amount,
        flags,
        0,
        eudx,
    )


def Trigger(  # noqa: N802
    players: list[Player],
    conditions: list[bytes] | bytes = [],
    actions: list[bytes] | bytes = [],
) -> bytes:
    conditions = FlattenList(conditions)
    actions = FlattenList(actions)

    ep_assert(isinstance(players, list), f"players is not list: {players}")
    ep_assert(isinstance(conditions, list), f"conditions is not list: {conditions}")
    ep_assert(isinstance(actions, list), f"actions is not list: {actions}")
    ep_assert(len(conditions) <= 16, f"16 < len(conditions): {len(conditions)}")
    ep_assert(len(actions) <= 64, f"64 < len(actions): {len(actions)}")
    peff = bytearray(28)
    for p in players:
        peff[EncodePlayer(p)] = 1

    b = b"".join(
        [
            *conditions,
            bytes(20 * (16 - len(conditions))),
            *actions,
            bytes(32 * (64 - len(actions))),
            b"\x04\0\0\0",
            bytes(peff),
        ]
    )
    ep_assert(len(b) == 2400, f"len(trigger) != 2400: {len(b)}")
    return b


# predefined conditions
def NoCondition() -> bytes:  # noqa: N802
    return Condition(0, 0, 0, 0, 0, 0, 0, 0)


def CountdownTimer(comparison: Comparison, time: int) -> bytes:  # noqa: N802
    comparison = EncodeComparison(comparison)
    return Condition(0, 0, time, 0, comparison, 1, 0, 0)


def Command(  # noqa: N802
    player: Player, comparison: Comparison, number: int, unit: Unit
) -> bytes:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(0, player, number, unit, comparison, 2, 0, 0)


def Bring(  # noqa: N802
    player: Player,
    comparison: Comparison,
    number: int,
    unit: Unit,
    location: Location,
) -> bytes:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    return Condition(location, player, number, unit, comparison, 3, 0, 0)


def Accumulate(  # noqa: N802
    player: Player,
    comparison: Comparison,
    number: int,
    resource_type: Resource,
) -> bytes:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    resource_type = EncodeResource(resource_type)
    return Condition(0, player, number, 0, comparison, 4, resource_type, 0)


def __kills__internal(
    player: Player, comparison: Comparison, number: int, unit: Unit
) -> bytes:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(0, player, number, unit, comparison, 5, 0, 0)


Kills = _Kills()
Kills._internalf = __kills__internal


def CommandMost(unit: Unit) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 6, 0, 0)


def CommandMostAt(unit: Unit, location: Location) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    return Condition(location, 0, 0, unit, 0, 7, 0, 0)


def MostKills(unit: Unit) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 8, 0, 0)


def HighestScore(score_type: _Score) -> bytes:  # noqa: N802
    score_type = EncodeScore(score_type)
    return Condition(0, 0, 0, 0, 0, 9, score_type, 0)


def MostResources(resource_type: Resource) -> bytes:  # noqa: N802
    resource_type = EncodeResource(resource_type)
    return Condition(0, 0, 0, 0, 0, 10, resource_type, 0)


def Switch(switch: _Switch, state: SwitchState) -> bytes:  # noqa: N802
    switch = EncodeSwitch(switch)
    state = EncodeSwitchState(state)
    return Condition(0, 0, 0, 0, state, 11, switch, 0)


def ElapsedTime(comparison: Comparison, time: int) -> bytes:  # noqa: N802
    comparison = EncodeComparison(comparison)
    return Condition(0, 0, time, 0, comparison, 12, 0, 0)


def Briefing() -> bytes:  # noqa: N802
    return Condition(0, 0, 0, 0, 0, 13, 0, 0)


def Opponents(player: Player, comparison: Comparison, number: int) -> bytes:  # noqa: N802
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    return Condition(0, player, number, 0, comparison, 14, 0, 0)


def Deaths(  # noqa: N802
    player: Player, comparison: Comparison, number: int, unit: Unit
) -> bytes:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(0, player, number, unit, comparison, 15, 0, 0)


def CommandLeast(unit: Unit) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 16, 0, 0)


def CommandLeastAt(unit: Unit, location: Location) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    return Condition(location, 0, 0, unit, 0, 17, 0, 0)


def LeastKills(unit: Unit) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    return Condition(0, 0, 0, unit, 0, 18, 0, 0)


def LowestScore(score_type: _Score) -> bytes:  # noqa: N802
    score_type = EncodeScore(score_type)
    return Condition(0, 0, 0, 0, 0, 19, score_type, 0)


def LeastResources(resource_type: Resource) -> bytes:  # noqa: N802
    resource_type = EncodeResource(resource_type)
    return Condition(0, 0, 0, 0, 0, 20, resource_type, 0)


def Score(  # noqa: N802
    player: Player, score_type: _Score, comparison: Comparison, number: int
) -> bytes:
    player = EncodePlayer(player)
    score_type = EncodeScore(score_type)
    comparison = EncodeComparison(comparison)
    return Condition(0, player, number, 0, comparison, 21, score_type, 0)


def Always() -> bytes:  # noqa: N802
    return Condition(0, 0, 0, 0, 0, 22, 0, 0)


def Never() -> bytes:  # noqa: N802
    return Condition(0, 0, 0, 0, 0, 23, 0, 0)


# predefined Action
def NoAction() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 0, 0, 4)


def Victory() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 1, 0, 4)


def Defeat() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 2, 0, 4)


def PreserveTrigger() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 3, 0, 4)


def Wait(time: int) -> bytes:  # noqa: N802
    return Action(0, 0, 0, time, 0, 0, 0, 4, 0, 4)


def PauseGame() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 5, 0, 4)


def UnpauseGame() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 6, 0, 4)


def Transmission(  # noqa: N802
    unit: Unit,
    where: Location,
    sound_name: String,
    time_modifier: Modifier,
    time: int,
    text: String,
    AlwaysDisplay: int = 4,  # noqa: N803
) -> bytes:
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    sound_name = EncodeString(sound_name)
    modifier = EncodeModifier(time_modifier)
    text = EncodeString(text)
    return Action(where, text, sound_name, time, 0, 0, unit, 7, modifier, 4)


def PlayWAV(sound_name: String) -> bytes:  # noqa: N802
    sound_name = EncodeString(sound_name)
    return Action(0, 0, sound_name, 0, 0, 0, 0, 8, 0, 4)


def DisplayText(text: String, AlwaysDisplay: int = 4) -> bytes:  # noqa: N802, N803
    text = EncodeString(text)
    return Action(0, text, 0, 0, 0, 0, 0, 9, 0, 4)


def CenterView(where: Location) -> bytes:  # noqa: N802
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, 0, 0, 0, 10, 0, 4)


def CreateUnitWithProperties(  # noqa: N802
    count: int,
    unit: Unit,
    where: Location,
    player: Player,
    properties: UnitProperty | bytes,
) -> bytes:
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    player = EncodePlayer(player)
    prop = EncodeProperty(properties)
    return Action(where, 0, 0, 0, player, prop, unit, 11, count, 28)


def SetMissionObjectives(text: String) -> bytes:  # noqa: N802
    text = EncodeString(text)
    return Action(0, text, 0, 0, 0, 0, 0, 12, 0, 4)


def SetSwitch(switch: _Switch, state: SwitchAction) -> bytes:  # noqa: N802
    switch = EncodeSwitch(switch)
    state = EncodeSwitchAction(state)
    return Action(0, 0, 0, 0, 0, switch, 0, 13, state, 4)


def SetCountdownTimer(time_modifier: Modifier, time: int) -> bytes:  # noqa: N802
    modifier = EncodeModifier(time_modifier)
    return Action(0, 0, 0, time, 0, 0, 0, 14, modifier, 4)


def RunAIScript(script: AIScript) -> bytes:  # noqa: N802
    script = EncodeAIScript(script)
    return Action(0, 0, 0, 0, 0, script, 0, 15, 0, 4)


def RunAIScriptAt(script: AIScript, where: Location) -> bytes:  # noqa: N802
    script = EncodeAIScript(script)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, 0, script, 0, 16, 0, 4)


def LeaderBoardControl(unit: Unit, label: String) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, unit, 17, 0, 20)


def LeaderBoardControlAt(  # noqa: N802
    unit: Unit, location: Location, label: String
) -> bytes:
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    label = EncodeString(label)
    return Action(location, label, 0, 0, 0, 0, unit, 18, 0, 20)


def LeaderBoardResources(resource_type: Resource, label: String) -> bytes:  # noqa: N802
    resource_type = EncodeResource(resource_type)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, resource_type, 19, 0, 4)


def LeaderBoardKills(unit: Unit, label: String) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, unit, 20, 0, 20)


def LeaderBoardScore(score_type: _Score, label: String) -> bytes:  # noqa: N802
    score_type = EncodeScore(score_type)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, score_type, 21, 0, 4)


def KillUnit(unit: Unit, player: Player) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    player = EncodePlayer(player)
    return Action(0, 0, 0, 0, player, 0, unit, 22, 0, 20)


def KillUnitAt(  # noqa: N802
    count: Count, unit: Unit, where: Location, for_player: Player
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    for_player = EncodePlayer(for_player)
    return Action(where, 0, 0, 0, for_player, 0, unit, 23, count, 20)


def RemoveUnit(unit: Unit, player: Player) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    player = EncodePlayer(player)
    return Action(0, 0, 0, 0, player, 0, unit, 24, 0, 20)


def RemoveUnitAt(  # noqa: N802
    count: Count, unit: Unit, where: Location, for_player: Player
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    for_player = EncodePlayer(for_player)
    return Action(where, 0, 0, 0, for_player, 0, unit, 25, count, 20)


def SetResources(  # noqa: N802
    player: Player, modifier: Modifier, amount: int, resource_type: Resource
) -> bytes:
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    resource_type = EncodeResource(resource_type)
    return Action(0, 0, 0, 0, player, amount, resource_type, 26, modifier, 4)


def SetScore(  # noqa: N802
    player: Player, modifier: Modifier, amount: int, score_type: _Score
) -> bytes:
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    score_type = EncodeScore(score_type)
    return Action(0, 0, 0, 0, player, amount, score_type, 27, modifier, 4)


def MinimapPing(where: Location) -> bytes:  # noqa: N802
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, 0, 0, 0, 28, 0, 4)


def TalkingPortrait(unit: Unit, time: int) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    return Action(0, 0, 0, time, 0, 0, unit, 29, 0, 20)


def MuteUnitSpeech() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 30, 0, 4)


def UnMuteUnitSpeech() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 31, 0, 4)


def LeaderBoardComputerPlayers(state: PropState) -> bytes:  # noqa: N802
    state = EncodePropState(state)
    return Action(0, 0, 0, 0, 0, 0, 0, 32, state, 4)


def LeaderBoardGoalControl(goal: int, unit: Unit, label: String) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, unit, 33, 0, 20)


def LeaderBoardGoalControlAt(  # noqa: N802
    goal: int, unit: Unit, location: Location, label: String
) -> bytes:
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    label = EncodeString(label)
    return Action(location, label, 0, 0, 0, goal, unit, 34, 0, 20)


def LeaderBoardGoalResources(  # noqa: N802
    goal: int, resource_type: Resource, label: String
) -> bytes:
    resource_type = EncodeResource(resource_type)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, resource_type, 35, 0, 4)


def LeaderBoardGoalKills(goal: int, unit: Unit, label: String) -> bytes:  # noqa: N802
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, unit, 36, 0, 20)


def LeaderBoardGoalScore(  # noqa: N802
    goal: int, score_type: _Score, label: String
) -> bytes:
    score_type = EncodeScore(score_type)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, score_type, 37, 0, 4)


def MoveLocation(  # noqa: N802
    location: Location, on_unit: Unit, owner: Player, dest_location: Location
) -> bytes:
    location = EncodeLocation(location)
    on_unit = EncodeUnit(on_unit)
    owner = EncodePlayer(owner)
    dest_location = EncodeLocation(dest_location)
    return Action(dest_location, 0, 0, 0, owner, location, on_unit, 38, 0, 20)


def MoveUnit(  # noqa: N802
    count: Count,
    unit_type: Unit,
    owner: Player,
    start_location: Location,
    dest_location: Location,
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit_type)
    owner = EncodePlayer(owner)
    startloc = EncodeLocation(start_location)
    dest_location = EncodeLocation(dest_location)
    return Action(startloc, 0, 0, 0, owner, dest_location, unit, 39, count, 20)


def LeaderBoardGreed(goal: int) -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, goal, 0, 40, 0, 4)


def SetNextScenario(scenario_name: String) -> bytes:  # noqa: N802
    scenario_name = EncodeString(scenario_name)
    return Action(0, scenario_name, 0, 0, 0, 0, 0, 41, 0, 4)


def SetDoodadState(  # noqa: N802
    state: PropState, unit: Unit, owner: Player, where: Location
) -> bytes:
    state = EncodePropState(state)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, 0, unit, 42, state, 20)


def SetInvincibility(  # noqa: N802
    state: PropState, unit: Unit, owner: Player, where: Location
) -> bytes:
    state = EncodePropState(state)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, 0, unit, 43, state, 20)


def CreateUnit(  # noqa: N802
    number: int, unit: Unit, where: Location, for_player: Player
) -> bytes:
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    for_player = EncodePlayer(for_player)
    return Action(where, 0, 0, 0, for_player, 0, unit, 44, number, 20)


def SetDeaths(  # noqa: N802
    player: Player, modifier: Modifier, number: int, unit: Unit
) -> bytes:
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    unit = EncodeUnit(unit)
    return Action(0, 0, 0, 0, player, number, unit, 45, modifier, 20)


def Order(  # noqa: N802
    unit: Unit,
    owner: Player,
    start_location: Location,
    order_type: _Order,
    dest_location: Location,
) -> bytes:
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    startloc = EncodeLocation(start_location)
    order_type = EncodeOrder(order_type)
    dest_location = EncodeLocation(dest_location)
    return Action(startloc, 0, 0, 0, owner, dest_location, unit, 46, order_type, 20)


def Comment(text: String) -> bytes:  # noqa: N802
    text = EncodeString(text)
    return Action(0, text, 0, 0, 0, 0, 0, 47, 0, 4)


def GiveUnits(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, new_owner: Player
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    new_owner = EncodePlayer(new_owner)
    return Action(where, 0, 0, 0, owner, new_owner, unit, 48, count, 20)


def ModifyUnitHitPoints(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, percent: int
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, percent, unit, 49, count, 20)


def ModifyUnitEnergy(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, percent: int
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, percent, unit, 50, count, 20)


def ModifyUnitShields(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, percent: int
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, percent, unit, 51, count, 20)


def ModifyUnitResourceAmount(  # noqa: N802
    count: Count, owner: Player, where: Location, new_value: int
) -> bytes:
    count = EncodeCount(count)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, new_value, 0, 52, count, 4)


def ModifyUnitHangarCount(  # noqa: N802
    add: int, count: Count, unit: Unit, owner: Player, where: Location
) -> bytes:
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, add, unit, 53, count, 20)


def PauseTimer() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 54, 0, 4)


def UnpauseTimer() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 55, 0, 4)


def Draw() -> bytes:  # noqa: N802
    return Action(0, 0, 0, 0, 0, 0, 0, 56, 0, 4)


def SetAllianceStatus(player: Player, status: AllyStatus) -> bytes:  # noqa: N802
    player = EncodePlayer(player)
    status = EncodeAllyStatus(status)
    return Action(0, 0, 0, 0, player, 0, status, 57, 0, 4)


# compound triggers
def Memory(dest: int, cmptype: Comparison, value: int) -> bytes:  # noqa: N802
    return Deaths(EPD(dest), cmptype, value, 0)


def SetMemory(dest: int, modtype, value: int) -> bytes:  # noqa: N802
    return SetDeaths(EPD(dest), modtype, value, 0)


def DeathsX(  # noqa: N802
    player: Player, comparison: Comparison, number: int, unit: Unit, mask: int
) -> bytes:
    player = EncodePlayer(player)
    comparison = EncodeComparison(comparison)
    unit = EncodeUnit(unit)
    return Condition(mask, player, number, unit, comparison, 15, 0, 0, eudx=0x4353)


def MemoryX(dest: int, cmptype: Comparison, value: int, mask: int) -> bytes:  # noqa: N802
    return DeathsX(EPD(dest), cmptype, value, 0, mask)


def SetDeathsX(  # noqa: N802
    player: Player, modifier: Modifier, number: int, unit: Unit, mask: int
) -> bytes:
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    unit = EncodeUnit(unit)
    return Action(mask, 0, 0, 0, player, number, unit, 45, modifier, 20, eudx=0x4353)


def SetMemoryX(dest: int, modtype: Modifier, value: int, mask: int) -> bytes:  # noqa: N802
    return SetDeathsX(EPD(dest), modtype, value, 0, mask)



from helper import *


# Basic struct test
class Coord(EUDStruct):
    _fields_ = ["x", "y"]


@TestInstance
def test_pool():
    from eudplib.collections.objpool import SetGlobalPoolFieldN

    # Basic allocation
    a = Coord.alloc()
    a.x = 5
    a.y = 7

    b = Coord.alloc()
    b.x = 9

    c = Coord.alloc()
    c.y = 11

    test_equality("Basic pool allocation", [a.x, a.y, b.x, c.y], [5, 7, 9, 11])

    # freeing
    Coord.free(a)
    a = None

    test_equality("After freeing", [b.x, c.y], [9, 11])

    # Reallocating some
    d = Coord.alloc()
    d.x = 3
    d.y = 6
    test_equality("Alloc after freeing", [d.x, d.y, b.x, c.y], [3, 6, 9, 11])



## NOTE: THIS FILE IS GENERATED BY EPSCRIPT! DO NOT MODITY
from eudplib import *
from eudplib.core.eudfunc import EUDTraceLog, EUDTracedFunc, EUDTracedTypedFunc, EUDTracedMethod, EUDTracedTypedMethod
from eudplib.epscript.helper import _RELIMP, _TYGV, _TYSV, _TYLV, _CGFW, _ARR, _VARR, _SRET, _SV, _ATTW, _ARRW, _ATTC, _ARRC, _L2V, _LSH, _ALL
# (Line 1) import .relimpCycle1;
# (Line 3) function updateUnitNameAndRank() {
try:
    relimpCycle1 = _RELIMP(".", "relimpCycle1")
except ImportError:
    from . import relimpCycle1
@EUDFunc
def f_updateUnitNameAndRank():
    # (Line 5) const NAME_STR = EncodeString(py_eval("99 * '\\r'"));
    NAME_STR = EncodeString(eval("99 * '\\r'"))
    # (Line 8) const RANK_TBL = $B("Recruit (Rank)");
    RANK_TBL = EncodeTBL("Recruit (Rank)")
    # (Line 10) const LOCAL_UNIT_SELECT = EPD(0x6284B8);  // 선택한 유닛 (비공유)
    LOCAL_UNIT_SELECT = EPD(0x6284B8)
    # (Line 11) /* TEST */ dwwrite_epd(LOCAL_UNIT_SELECT, 0x59CCA8 + 336);
    f_dwwrite_epd(LOCAL_UNIT_SELECT, 0x59CCA8 + 336)
    # (Line 12) static var selectedUnitPtr, selectedUnitEpd = 0, 0;
    selectedUnitPtr, selectedUnitEpd = _TYSV([None, None], [0, 0])
    # (Line 13) if (!MemoryEPD(LOCAL_UNIT_SELECT, Exactly, selectedUnitPtr)) {
    if EUDIf()(MemoryEPD(LOCAL_UNIT_SELECT, Exactly, selectedUnitPtr), neg=True):
        # (Line 14) selectedUnitPtr, selectedUnitEpd = cunitepdread_epd(LOCAL_UNIT_SELECT);
        _SV([selectedUnitPtr, selectedUnitEpd], [f_cunitepdread_epd(LOCAL_UNIT_SELECT)])
        # (Line 15) }
        # (Line 16) if (selectedUnitPtr == 0) return -1, -1;
    EUDEndIf()
    if EUDIf()(selectedUnitPtr == 0):
        EUDReturn(-1, -1)
        # (Line 17) /* TEST */ dwwrite_epd(LOCAL_UNIT_SELECT, 0);
    EUDEndIf()
    f_dwwrite_epd(LOCAL_UNIT_SELECT, 0)
    # (Line 18) const namePtr = GetMapStringAddr(NAME_STR);
    namePtr = GetMapStringAddr(NAME_STR)
    # (Line 19) var rankPtr;
    rankPtr = EUDVariable()
    # (Line 20) once rankPtr = GetTBLAddr(RANK_TBL);
    _t3 = EUDExecuteOnce()
    if _t3():
        rankPtr << (GetTBLAddr(RANK_TBL))
        # (Line 21) const unitTypes = py_list();
    EUDEndExecuteOnce()
    unitTypes = list()
    # (Line 22) const unit = CUnit(selectedUnitEpd, ptr=selectedUnitPtr);
    unit = CUnit(selectedUnitEpd, ptr=selectedUnitPtr)
    # (Line 24) const Ticket = EUDArray(py_range(8));
    Ticket = EUDArray(range(8))
    # (Line 25) const Ore = EUDArray(py_range(1, 17, 2));
    Ore = EUDArray(range(1, 17, 2))
    # (Line 26) const Gas = EUDArray(py_range(2, 26, 3));
    Gas = EUDArray(range(2, 26, 3))
    # (Line 28) /* TEST */ dwwrite(0x59CCA8 + 336 + 0x64, $U("Terran Academy"));
    f_dwwrite(0x59CCA8 + 336 + 0x64, EncodeUnit("Terran Academy"))
    # (Line 29) /* TEST */ bwrite(0x59CCA8 + 336 + 0x4C, $P2);
    f_bwrite(0x59CCA8 + 336 + 0x4C, 1)
    # (Line 31) epdswitch (unit + 0x64/4, 255) {
    EPDSwitch(unit + 0x64 // 4, 255)
    # (Line 32) case $U("Terran Academy"):
    _t4 = EUDSwitchCase()
    # (Line 33) sprintf(namePtr, "Ticket {}, Ore {}, Gas {}", Ticket[unit.owner], Ore[unit.owner], Gas[unit.owner]);
    if _t4(EncodeUnit("Terran Academy")):
        f_sprintf(namePtr, "Ticket {}, Ore {}, Gas {}", Ticket[unit.owner], Ore[unit.owner], Gas[unit.owner])
        # (Line 34) sprintf(rankPtr, "{}", unit.unitType);
        f_sprintf(rankPtr, "{}", unit.unitType)
        # (Line 35) break;
        EUDBreak()
        # (Line 36) unitTypes.extend((EUDGetLastBlockOfName("swblock"))[1][py_str("casebrlist")].keys());
        unitTypes.extend((EUDGetLastBlockOfName("swblock"))[1][str("casebrlist")].keys())
        # (Line 37) }
    # (Line 38) once {
    EUDEndSwitch()
    _t5 = EUDExecuteOnce()
    if _t5():
        # (Line 39) const actions = py_list();
        actions = list()
        # (Line 40) foreach(unitType : unitTypes) {
        for unitType in unitTypes:
            # (Line 41) actions.append(SetMemoryX(
            # (Line 42) 0x660260 + 2 * unitType,
            # (Line 43) SetTo,
            # (Line 44) NAME_STR << (16 * (unitType % 2)),
            # (Line 45) 0xFFFF << (16 * (unitType % 2))
            # (Line 46) ));
            actions.append(SetMemoryX(0x660260 + 2 * unitType, SetTo, _LSH(NAME_STR,(16 * (unitType % 2))), _LSH(0xFFFF,(16 * (unitType % 2)))))
            # (Line 47) }
            # (Line 48) DoActions(actions);

        DoActions(actions)
        # (Line 49) }
        # (Line 51) setcurpl(getuserplayerid());
    EUDEndExecuteOnce()
    f_setcurpl(f_getuserplayerid())
    # (Line 52) DisplayText(NAME_STR);
    # (Line 53) const tests = "Ticket 1, Ore 3, Gas 5", "{}".format($U("Terran Academy"));
    DoActions(DisplayText(NAME_STR))
    tests = "Ticket 1, Ore 3, Gas 5", "{}".format(EncodeUnit("Terran Academy"))
    # (Line 54) return memcmp(namePtr, Db(tests[0]), py_len(tests[0].encode("utf-8")) + 1),
    # (Line 55) memcmp(rankPtr, Db(tests[1]), py_len(tests[1].encode("utf-8")) + 1);
    EUDReturn(f_memcmp(namePtr, Db(tests[0]), len(tests[0].encode("utf-8")) + 1), f_memcmp(rankPtr, Db(tests[1]), len(tests[1].encode("utf-8")) + 1))
    # (Line 56) }



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...utils import EPD, i2b4
from .. import rawtrigger as bt
from ..eudobj import Db
from .vbase import VariableBase


class EUDLightVariable(VariableBase):
    __slots__ = ("_memaddr",)

    def __init__(self, initvalue=0, /):
        super().__init__()
        self._memaddr = Db(i2b4(initvalue))

    def getValueAddr(self):  # noqa: N802
        return self._memaddr

    def checkNonRValue(self):  # noqa: N802
        pass

    def __hash__(self):
        return id(self)


class EUDLightBool:
    __slots__ = ("_mask", "_memaddr", "_memepd")
    _init_true_lv = EUDLightVariable(0xFFFFFFFF)
    _init_true_bit = 0
    _init_false_lv = EUDLightVariable(0)
    _init_false_bit = 0

    def __init__(self, initval: bool = False, /):
        if initval:
            if EUDLightBool._init_true_bit >= 32:
                EUDLightBool._init_true_lv = EUDLightVariable(0xFFFFFFFF)
                EUDLightBool._init_true_bit = 0
            lv = EUDLightBool._init_true_lv
            bit = EUDLightBool._init_true_bit
            EUDLightBool._init_true_bit += 1
        else:
            if EUDLightBool._init_false_bit >= 32:
                EUDLightBool._init_false_lv = EUDLightVariable(0)
                EUDLightBool._init_false_bit = 0
            lv = EUDLightBool._init_false_lv
            bit = EUDLightBool._init_false_bit
            EUDLightBool._init_false_bit += 1

        self._mask = 1 << bit
        self._memaddr = lv._memaddr
        self._memepd = EPD(lv._memaddr)

    def getValueAddr(self):  # noqa: N802
        return self._memaddr

    def Set(self):  # noqa: N802
        return bt.SetMemoryXEPD(self._memepd, bt.SetTo, self._mask, self._mask)

    def Clear(self):  # noqa: N802
        return bt.SetMemoryXEPD(self._memepd, bt.SetTo, 0, self._mask)

    def Toggle(self):  # noqa: N802
        return bt.SetMemoryXEPD(self._memepd, bt.Add, self._mask, self._mask)

    def IsSet(self):  # noqa: N802
        return bt.MemoryXEPD(self._memepd, bt.AtLeast, 1, self._mask)

    def IsCleared(self):  # noqa: N802
        return bt.MemoryXEPD(self._memepd, bt.Exactly, 0, self._mask)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .chktok import CHK


class PlayerInfo:
    def __init__(self) -> None:
        self.force: int
        self.race: int
        self.type: int
        self.racestr: str
        self.typestr: str


_playerinfo: list[PlayerInfo] = []


def init_player_info(chkt: CHK) -> None:
    _playerinfo.clear()

    section_forc = chkt.getsection("FORC")
    section_ownr = chkt.getsection("OWNR")
    section_side = chkt.getsection("SIDE")

    for player in range(8):
        pinfo = PlayerInfo()
        _playerinfo.append(pinfo)

        # Numeric info
        pinfo.force = section_forc[player]
        pinfo.race = section_side[player]
        pinfo.type = section_ownr[player]

        # String info
        pinfo.racestr = {
            0x00: "Zerg",
            0x01: "Terran",
            0x02: "Protoss",
            0x03: "Independent",
            0x04: "Neutral",
            0x05: "User selectable",
            0x07: "Inactive",
            0x08: "Human",
            0x0A: "Human",
            0x0D: "Starcraft Campaign Editor",
        }.get(pinfo.race, "")

        pinfo.typestr = {
            0x00: "Unused",
            0x03: "Rescuable",
            0x05: "Computer",
            0x06: "Human",
            0x07: "Neutral",
        }.get(pinfo.type, "")


def GetPlayerInfo(player) -> PlayerInfo:  # noqa: N802
    from ..rawtrigger.constenc import EncodePlayer

    return _playerinfo[EncodePlayer(player)]



# Copyright 2019 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from .. import utils as ut


class EUDVArrayReader:
    def __init__(self):
        self._trg = c.Forward()
        self._fin = c.Forward()

    def _maketrg(self):
        c.PushTriggerScope()
        self._trg << c.RawTrigger(
            nextptr=0,
            actions=[
                c.SetDeaths(0, c.SetTo, 0, 0),
                c.SetNextPtr(0, self._fin),
            ],
        )
        self._fin << c.RawTrigger(
            nextptr=0,
            actions=[
                c.SetMemory(self._trg + 4, c.Add, 72),
                c.SetMemory(self._trg + 328 + 16, c.Add, 18),
                c.SetMemory(self._trg + 360 + 16, c.Add, 18),
            ],
        )
        c.PopTriggerScope()

    def seek(self, varr_ptr, varr_epd, eudv, acts=[]):
        if not self._trg.IsSet():
            self._maketrg()

        if c.IsEUDVariable(varr_ptr):
            c.VProc(
                [varr_ptr, varr_epd],
                [
                    varr_ptr.QueueAssignTo(ut.EPD(self._trg) + 1),
                    varr_epd.AddNumber(1),
                    varr_epd.QueueAssignTo(ut.EPD(self._trg) + 360 // 4 + 4),
                    c.SetMemory(
                        self._trg + 328 + 20,
                        c.SetTo,
                        ut.EPD(eudv.getValueAddr()),
                    ),
                ],
            )
            trg = c.VProc(
                varr_epd,
                [acts, varr_epd.AddNumber(328 // 4 + 3), varr_epd.AddDest(-8)],
            )
            return trg
        else:
            return [
                c.SetNextPtr(self._trg, varr_ptr),
                c.SetMemory(self._trg + 328 + 16, c.SetTo, varr_epd + 328 // 4 + 4),
                c.SetMemory(
                    self._trg + 328 + 20, c.SetTo, ut.EPD(eudv.getValueAddr())
                ),
                c.SetMemory(self._trg + 360 + 16, c.SetTo, varr_epd + 1),
            ]

    def read(self, acts=[]):
        if not self._trg.IsSet():
            self._maketrg()
        nptr = c.Forward()
        trg = c.RawTrigger(
            nextptr=self._trg, actions=[acts, c.SetNextPtr(self._fin, nptr)]
        )
        nptr << c.NextTrigger()
        return trg



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
from ...core import Db, SetMemory, SetTo
from ...ctrlstru import DoActions
from ...memio import f_epdread_epd, f_repmovsd_epd
from ...utils import EPD
from .mempatch import f_dwpatch_epd

_data = []


def _add_datadumper(input_data, out_offsets, flags):
    _data.append((input_data, out_offsets, flags))


def _f_datadumper():
    for input_data, out_offsets, flags in _data:
        if len(out_offsets) == 0:
            continue

        # Reset?
        if "unpatchable" in flags:
            assert "copy" not in flags, "Cannot apply both 'copy' and 'unpatchable'"
            for out_offset in out_offsets:
                f_dwpatch_epd(EPD(out_offset), Db(input_data))

        elif "copy" in flags:
            input_data_db = Db(input_data)
            input_dwordn = (len(input_data) + 3) // 4

            for out_offset in out_offsets:
                addr_epd = f_epdread_epd(EPD(out_offset))
                f_repmovsd_epd(addr_epd, EPD(input_data_db), input_dwordn)

        else:
            DoActions(
                [
                    SetMemory(out_offset, SetTo, Db(input_data))
                    for out_offset in out_offsets
                ]
            )



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from .. import core as c
from ..trigger.tpatcher import negate_cond
from ..utils.eperror import EPError
from .simpleblock import EUDElse, EUDElseIf, EUDElseIfNot, EUDEndIf, EUDIf, EUDIfNot


def EUDOr(cond1, *conds) -> c.EUDLightBool:  # noqa: N802
    """cond1 || cond2 || ... || condn

    .. warning:: Short circuiting is not supported

    :param conds: List of conditions
    """

    v = c.EUDLightBool()
    if EUDIf()(cond1):
        c.RawTrigger(actions=v.Set())
    for cond in conds:
        if EUDElseIf()(cond):
            c.RawTrigger(actions=v.Set())
    if EUDElse()():
        c.RawTrigger(actions=v.Clear())
    EUDEndIf()
    return v


def EUDAnd(cond1, *conds) -> c.EUDLightBool:  # noqa: N802
    """cond1 && cond2 && ... && condn

    .. note::
        This function computes AND value of various conditions.
        If you don't want to do much computation, you should better use
        plain list instead of this function.

    .. warning:: Short circuiting is not supported.

    :param conds: List of conditions
    """

    v = c.EUDLightBool()
    if EUDIfNot()(cond1):
        c.RawTrigger(actions=v.Clear())
    for cond in conds:
        if EUDElseIfNot()(cond):
            c.RawTrigger(actions=v.Clear())
    if EUDElse()():
        c.RawTrigger(actions=v.Set())
    EUDEndIf()
    return v


def EUDNot(cond):  # noqa: N802
    """!cond

    :param conds: Condition to negate
    """

    if c.IsEUDVariable(cond):
        # TODO: better handle !v in conditional
        v = c.EUDVariable()
        v << 0
        c.RawTrigger(conditions=cond.Exactly(0), actions=v.SetNumber(1))
        return v

    if isinstance(cond, bool):
        if cond:
            return False
        return True
    try:
        return negate_cond(cond)
    except (AttributeError, EPError):
        pass
    v = c.EUDLightBool()
    c.RawTrigger(actions=v.Set())
    if EUDIf()(cond):
        c.RawTrigger(actions=v.Clear())
    EUDEndIf()
    return v
